import dataclasses
from typing import Dict, Final, ClassVar, Any, List, Optional

from dlt.common.destination.configuration import CsvFormatConfiguration
from dlt.common.configuration import configspec
from dlt.common.configuration.specs import ConnectionStringCredentials
from dlt.common.utils import digest128
from dlt.common.typing import TSecretStrValue

from dlt.common.destination.client import DestinationClientDwhWithStagingConfiguration


@configspec(init=False)
class PostgresCredentials(ConnectionStringCredentials):
    drivername: Final[str] = dataclasses.field(default="postgresql", init=False, repr=False, compare=False)  # type: ignore
    database: str = None
    username: str = None
    password: TSecretStrValue = None
    host: str = None
    port: int = 5432
    connect_timeout: int = 15
    client_encoding: Optional[str] = None

    __config_gen_annotations__: ClassVar[List[str]] = ["port", "connect_timeout"]

    def parse_native_representation(self, native_value: Any) -> None:
        super().parse_native_representation(native_value)
        self.connect_timeout = int(self.query.get("connect_timeout", self.connect_timeout))
        self.client_encoding = self.query.get("client_encoding", self.client_encoding)

    def get_query(self) -> Dict[str, Any]:
        query = dict(super().get_query())
        query["connect_timeout"] = self.connect_timeout
        if self.client_encoding:
            query["client_encoding"] = self.client_encoding
        return query


@configspec
class PostgresClientConfiguration(DestinationClientDwhWithStagingConfiguration):
    destination_type: Final[str] = dataclasses.field(default="postgres", init=False, repr=False, compare=False)  # type: ignore
    credentials: PostgresCredentials = None

    create_indexes: bool = True

    csv_format: Optional[CsvFormatConfiguration] = None
    """Optional csv format configuration"""

    def fingerprint(self) -> str:
        """Returns a fingerprint of host part of a connection string"""
        if self.credentials and self.credentials.host:
            return digest128(self.credentials.host)
        return ""



import dataclasses

from typing import ClassVar, List

from dlt.common.configuration import configspec
from dlt.common.configuration.specs import BaseConfiguration, known_sections

from dlt._workspace.helpers.dashboard.const import (
    DEFAULT_COLUMN_OTHER_HINTS,
    DEFAULT_COLUMN_TYPE_HINTS,
    DEFAULT_DATETIME_FORMAT,
    DEFAULT_TABLE_LIST_FIELDS,
)


@configspec
class DashboardConfiguration(BaseConfiguration):
    table_list_fields: List[str] = dataclasses.field(
        default_factory=lambda: list(DEFAULT_TABLE_LIST_FIELDS)
    )
    """The fields to show in the table lists, name is always present and cannot be removed"""

    column_type_hints: List[str] = dataclasses.field(
        default_factory=lambda: list(DEFAULT_COLUMN_TYPE_HINTS)
    )
    """Which column hints to show in the column list if type hints are enabled"""

    column_other_hints: List[str] = dataclasses.field(
        default_factory=lambda: list(DEFAULT_COLUMN_OTHER_HINTS)
    )
    """Which column hints to show in the column list if other hints are enabled"""

    datetime_format: str = DEFAULT_DATETIME_FORMAT
    """The format of the datetime strings"""

    sync_from_runtime: bool = False
    """
    Whether to sync the pipeline states and traces from the runtime backup.
    Needs to be run inside a dlt workspace with runtime artifacts credentials set.
    """

    # this is needed for using this as a param in the cache
    def __hash__(self) -> int:
        return hash(
            ", ".join(self.table_list_fields)
            + ", ".join(self.column_type_hints)
            + ", ".join(self.column_other_hints)
            + self.datetime_format
        )

    __section__: ClassVar[str] = known_sections.DASHBOARD



import os

from dlt._workspace._workspace_context import WorkspaceRunContext, active
from dlt._workspace.profile import (
    get_profile_pin_file,
    read_profile_pin,
    save_profile_pin,
)
from dlt._workspace.cli import echo as fmt, utils
from dlt._workspace.cli.utils import fetch_profiles_list


@utils.track_command("profile", track_before=False, operation="info")
def print_profile_info(workspace_run_context: WorkspaceRunContext) -> None:
    fmt.echo("Current profile: %s" % fmt.bold(workspace_run_context.profile))
    if pinned_profile := read_profile_pin(workspace_run_context):
        fmt.echo("Pinned profile: %s" % fmt.bold(pinned_profile))


@utils.track_command("profile", track_before=False, operation="list")
def list_profiles(workspace_run_context: WorkspaceRunContext) -> None:
    fmt.echo("Available profiles:")
    for p in fetch_profiles_list():
        markers = []
        if p["is_current"]:
            markers.append(fmt.bold("(current)"))
        if p["is_configured"]:
            markers.append(fmt.bold("(configured)"))
        marker_str = " ".join(markers)
        if marker_str:
            fmt.echo("* %s %s - %s" % (fmt.bold(p["name"]), marker_str, p["description"]))
        else:
            fmt.echo("* %s - %s" % (fmt.bold(p["name"]), p["description"]))


@utils.track_command("profile", track_before=False, operation="pin")
def pin_profile(workspace_run_context: WorkspaceRunContext, profile_name: str) -> None:
    if not profile_name:
        pinned_profile = read_profile_pin(workspace_run_context)
        if pinned_profile:
            pin_file = get_profile_pin_file(workspace_run_context)
            fmt.echo(
                "Currently pinned profile is: %s. To unpin remove %s file."
                % (fmt.bold(pinned_profile), fmt.bold(os.path.relpath(pin_file)))
            )
        else:
            fmt.echo("No pinned profile.")
    else:
        fmt.echo("Will pin the profile %s to current Workspace." % fmt.bold(profile_name))
        save_profile_pin(workspace_run_context, profile_name)



import os
from abc import ABC, abstractmethod
import dataclasses
import datetime  # noqa: 251
import warnings
import humanize
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generic,
    List,
    NamedTuple,
    Optional,
    Protocol,
    Sequence,
    Tuple,
    TypeVar,
    Mapping,
    Literal,
    Union,
    Mapping,
)
from typing_extensions import NotRequired

from dlt.common.configuration.specs.pluggable_run_context import RunContextBase
from dlt.common.typing import TypedDict
from dlt.common.configuration import configspec
from dlt.common.configuration.container import Container
from dlt.common.configuration.exceptions import ContextDefaultCannotBeCreated
from dlt.common.configuration.specs import ContainerInjectableContext
from dlt.common.configuration.specs import RuntimeConfiguration
from dlt.common.destination import TDestinationReferenceArg, AnyDestination
from dlt.common.destination.client import JobClientBase
from dlt.common.destination.exceptions import DestinationHasFailedJobs
from dlt.common.metrics import (
    DataWriterMetrics,
    ExtractDataInfo,
    ExtractMetrics,
    LoadMetrics,
    NormalizeMetrics,
    StepMetrics,
    DataWriterAndCustomMetrics,
)
from dlt.common.schema import Schema
from dlt.common.schema.typing import (
    TColumnSchema,
    TWriteDispositionConfig,
    TSchemaContract,
)
from dlt.common.storages.load_package import ParsedLoadJobFileName
from dlt.common.storages.load_storage import LoadPackageInfo
from dlt.common.time import ensure_pendulum_datetime_utc, precise_time
from dlt.common.typing import DictStrAny, StrAny, SupportsHumanize, TColumnNames
from dlt.common.data_writers.writers import TLoaderFileFormat
from dlt.common.utils import RowCounts, merge_row_counts
from dlt.common.versioned_state import TVersionedState
from dlt.common.runtime.collector_base import Collector


# TRefreshMode = Literal["full", "replace"]
TRefreshMode = Literal["drop_sources", "drop_resources", "drop_data"]


class _StepInfo(NamedTuple):
    pipeline: "SupportsPipeline"
    loads_ids: List[str]
    """ids of the loaded packages"""
    load_packages: List[LoadPackageInfo]
    """Information on loaded packages"""
    first_run: bool
    started_at: datetime.datetime
    finished_at: datetime.datetime


TStepMetricsCo = TypeVar("TStepMetricsCo", bound=StepMetrics, covariant=True)


class StepInfo(SupportsHumanize, Generic[TStepMetricsCo]):
    pipeline: "SupportsPipeline"
    metrics: Dict[str, List[TStepMetricsCo]]
    """Metrics per load id. If many sources with the same name were extracted, there will be more than 1 element in the list"""
    loads_ids: List[str]
    """ids of the loaded packages"""
    load_packages: List[LoadPackageInfo]
    """Information on loaded packages"""
    first_run: bool

    @property
    def started_at(self) -> datetime.datetime:
        """Returns the earliest start date of all collected metrics"""
        if not self.metrics:
            return None
        try:
            return min(m["started_at"] for l_m in self.metrics.values() for m in l_m)
        except ValueError:
            return None

    @property
    def finished_at(self) -> datetime.datetime:
        """Returns the latest end date of all collected metrics"""
        if not self.metrics:
            return None
        try:
            return max(m["finished_at"] for l_m in self.metrics.values() for m in l_m)
        except ValueError:
            return None

    @property
    def is_empty(self) -> bool:
        """Returns True if step didn't process any load packages."""
        return bool(self.loads_ids) is False

    def asdict(self) -> DictStrAny:
        # to be mixed with NamedTuple
        step_info: DictStrAny = self._asdict()  # type: ignore
        step_info["pipeline"] = {"pipeline_name": self.pipeline.pipeline_name}
        step_info["load_packages"] = [package.asdict() for package in self.load_packages]
        if self.metrics:
            step_info["started_at"] = self.started_at
            step_info["finished_at"] = self.finished_at
            all_metrics = []
            for load_id, metrics in step_info["metrics"].items():
                for metric in metrics:
                    all_metrics.append({**dict(metric), "load_id": load_id})

            step_info["metrics"] = all_metrics
        return step_info

    def __str__(self) -> str:
        return self.asstr(verbosity=0)

    @staticmethod
    def _load_packages_asstr(load_packages: List[LoadPackageInfo], verbosity: int) -> str:
        msg: str = ""
        for load_package in load_packages:
            cstr = (
                load_package.state.upper()
                if load_package.completed_at
                else f"{load_package.state.upper()} and NOT YET LOADED to the destination"
            )
            # now enumerate all complete loads if we have any failed packages
            # complete but failed job will not raise any exceptions
            failed_jobs = load_package.jobs["failed_jobs"]
            jobs_str = "no failed jobs" if not failed_jobs else f"{len(failed_jobs)} FAILED job(s)!"
            msg += f"\nLoad package {load_package.load_id} is {cstr} and contains {jobs_str}"
            if verbosity > 0:
                for failed_job in failed_jobs:
                    msg += (
                        f"\n\t[{failed_job.job_file_info.job_id()}]: {failed_job.failed_message}\n"
                    )
            if verbosity > 1:
                msg += "\nPackage details:\n"
                msg += load_package.asstr() + "\n"
        return msg

    @staticmethod
    def writer_metrics_asdict(
        job_metrics: Mapping[str, Union[DataWriterMetrics, DataWriterAndCustomMetrics]],
        key_name: str = "job_id",
        extend: StrAny = None,
    ) -> List[DictStrAny]:
        entities = []
        for entity_id, metrics in job_metrics.items():
            d = metrics._asdict()
            if extend:
                d.update(extend)
            d[key_name] = entity_id
            # add job-level info if known
            if metrics.file_path:
                d["table_name"] = ParsedLoadJobFileName.parse(metrics.file_path).table_name
            entities.append(d)
        return entities

    def _astuple(self) -> _StepInfo:
        return _StepInfo(
            self.pipeline,
            self.loads_ids,
            self.load_packages,
            self.first_run,
            self.started_at,
            self.finished_at,
        )


class _ExtractInfo(NamedTuple):
    """NamedTuple cannot be part of the derivation chain so we must re-declare all fields to use it as mixin later"""

    pipeline: "SupportsPipeline"
    metrics: Dict[str, List[ExtractMetrics]]
    extract_data_info: List[ExtractDataInfo]
    loads_ids: List[str]
    """ids of the loaded packages"""
    load_packages: List[LoadPackageInfo]
    """Information on loaded packages"""
    first_run: bool


class ExtractInfo(StepInfo[ExtractMetrics], _ExtractInfo):  # type: ignore[misc]
    """A tuple holding information on extracted data items. Returned by pipeline `extract` method."""

    def asdict(self) -> DictStrAny:
        """A dictionary representation of ExtractInfo that can be loaded with `dlt`"""
        d = super().asdict()
        d.pop("extract_data_info")
        # transform metrics
        d.pop("metrics")
        load_metrics: Dict[str, List[Any]] = {
            "job_metrics": [],
            "table_metrics": [],
            "resource_metrics": [],
            "dag": [],
            "hints": [],
        }
        for load_id, metrics_list in self.metrics.items():
            for idx, metrics in enumerate(metrics_list):
                extend = {"load_id": load_id, "extract_idx": idx}
                load_metrics["resource_metrics"].extend(
                    self.writer_metrics_asdict(
                        metrics["resource_metrics"], key_name="resource_name", extend=extend
                    )
                )
                load_metrics["dag"].extend(
                    [
                        {**extend, "parent_name": edge[0], "resource_name": edge[1]}
                        for edge in metrics["dag"]
                    ]
                )
                load_metrics["hints"].extend(
                    [
                        {**extend, "resource_name": name, **hints}
                        for name, hints in metrics["hints"].items()
                    ]
                )
                load_metrics["job_metrics"].extend(
                    self.writer_metrics_asdict(metrics["job_metrics"], extend=extend)
                )
                load_metrics["table_metrics"].extend(
                    self.writer_metrics_asdict(
                        metrics["table_metrics"], key_name="table_name", extend=extend
                    )
                )

        d.update(load_metrics)
        return d

    def asstr(self, verbosity: int = 0) -> str:
        return self._load_packages_asstr(self.load_packages, verbosity)


class _NormalizeInfo(NamedTuple):
    pipeline: "SupportsPipeline"
    metrics: Dict[str, List[NormalizeMetrics]]
    loads_ids: List[str]
    """ids of the loaded packages"""
    load_packages: List[LoadPackageInfo]
    """Information on loaded packages"""
    first_run: bool


class NormalizeInfo(StepInfo[NormalizeMetrics], _NormalizeInfo):  # type: ignore[misc]
    """A tuple holding information on normalized data items. Returned by pipeline `normalize` method."""

    @property
    def row_counts(self) -> RowCounts:
        if not self.metrics:
            return {}
        counts: RowCounts = {}
        for metrics in self.metrics.values():
            assert len(metrics) == 1, "Cannot deal with more than 1 normalize metric per load_id"
            merge_row_counts(
                counts, {t: m.items_count for t, m in metrics[0]["table_metrics"].items()}
            )
        return counts

    def asdict(self) -> DictStrAny:
        """A dictionary representation of NormalizeInfo that can be loaded with `dlt`"""
        d = super().asdict()
        # transform metrics
        d.pop("metrics")
        load_metrics: Dict[str, List[Any]] = {
            "job_metrics": [],
            "table_metrics": [],
        }
        for load_id, metrics_list in self.metrics.items():
            for idx, metrics in enumerate(metrics_list):
                extend = {"load_id": load_id, "extract_idx": idx}
                load_metrics["job_metrics"].extend(
                    self.writer_metrics_asdict(metrics["job_metrics"], extend=extend)
                )
                load_metrics["table_metrics"].extend(
                    self.writer_metrics_asdict(
                        metrics["table_metrics"], key_name="table_name", extend=extend
                    )
                )
        d.update(load_metrics)
        return d

    def asstr(self, verbosity: int = 0) -> str:
        if self.row_counts:
            msg = "Normalized data for the following tables:\n"
            for key, value in self.row_counts.items():
                msg += f"- {key}: {value} row(s)\n"
        else:
            msg = "No data found to normalize"
        msg += self._load_packages_asstr(self.load_packages, verbosity)
        return msg


class _LoadInfo(NamedTuple):
    pipeline: "SupportsPipeline"
    metrics: Dict[str, List[LoadMetrics]]
    destination_type: str
    destination_displayable_credentials: str
    destination_name: str
    environment: str
    staging_type: str
    staging_name: str
    staging_displayable_credentials: str
    destination_fingerprint: str
    dataset_name: str
    loads_ids: List[str]
    """ids of the loaded packages"""
    load_packages: List[LoadPackageInfo]
    """Information on loaded packages"""
    first_run: bool


class LoadInfo(StepInfo[LoadMetrics], _LoadInfo):  # type: ignore[misc]
    """A tuple holding the information on recently loaded packages. Returned by pipeline `run` and `load` methods"""

    def asdict(self) -> DictStrAny:
        """A dictionary representation of LoadInfo that can be loaded with `dlt`"""
        d = super().asdict()
        # transform metrics
        d.pop("metrics")
        load_metrics: Dict[str, List[Any]] = {"job_metrics": []}
        for load_id, metrics_list in self.metrics.items():
            # one set of metrics per package id
            assert len(metrics_list) == 1
            metrics = metrics_list[0]
            for job_metrics in metrics["job_metrics"].values():
                load_metrics["job_metrics"].append({"load_id": load_id, **job_metrics._asdict()})

        d.update(load_metrics)
        return d

    def asstr(self, verbosity: int = 0) -> str:
        msg = f"Pipeline {self.pipeline.pipeline_name} load step completed in "
        if self.started_at:
            elapsed = self.finished_at - self.started_at
            msg += humanize.precisedelta(elapsed)
        else:
            msg += "---"
        msg += (
            f"\n{len(self.loads_ids)} load package(s) were loaded to destination"
            f" {self.destination_name} and into dataset {self.dataset_name}\n"
        )
        if self.staging_name:
            msg += (
                f"The {self.staging_name} staging destination used"
                f" {self.staging_displayable_credentials} location to stage data\n"
            )

        msg += (
            f"The {self.destination_name} destination used"
            f" {self.destination_displayable_credentials} location to store data"
        )
        msg += self._load_packages_asstr(self.load_packages, verbosity)

        return msg

    @property
    def has_failed_jobs(self) -> bool:
        """Returns True if any of the load packages has a failed job."""
        for load_package in self.load_packages:
            if len(load_package.jobs["failed_jobs"]):
                return True
        return False

    def raise_on_failed_jobs(self) -> None:
        """Raises `DestinationHasFailedJobs` exception if any of the load packages has a failed job."""
        for load_package in self.load_packages:
            failed_jobs = load_package.jobs["failed_jobs"]
            if len(failed_jobs):
                raise DestinationHasFailedJobs(
                    self.destination_name, load_package.load_id, failed_jobs
                )

    def __str__(self) -> str:
        return self.asstr(verbosity=1)


TStepMetrics = TypeVar("TStepMetrics", bound=StepMetrics, covariant=False)
TStepInfo = TypeVar("TStepInfo", bound=StepInfo[StepMetrics])


class WithStepInfo(ABC, Generic[TStepMetrics, TStepInfo]):
    """Implemented by classes that generate StepInfo with metrics and package infos"""

    _current_load_id: str
    _load_id_metrics: Dict[str, List[TStepMetrics]]
    _current_load_started: float
    """Completed load ids metrics"""

    def __init__(self) -> None:
        self._load_id_metrics = {}
        self._current_load_id = None
        self._current_load_started = None

    def _step_info_start_load_id(self, load_id: str) -> None:
        self._current_load_id = load_id
        self._current_load_started = precise_time()
        self._load_id_metrics.setdefault(load_id, [])

    def _step_info_update_metrics(
        self, load_id: str, metrics: TStepMetrics, immutable: bool = False
    ) -> None:
        metrics["started_at"] = ensure_pendulum_datetime_utc(self._current_load_started)
        step_metrics = self._load_id_metrics[load_id]
        if immutable or len(step_metrics) == 0:
            step_metrics.append(metrics)
        else:
            step_metrics[0] = metrics

    def _step_info_complete_load_id(self, load_id: str, finished: bool = True) -> None:
        assert self._current_load_id == load_id, (
            f"Current load id mismatch {self._current_load_id} != {load_id} when completing step"
            " info"
        )
        # metrics must be present
        metrics = self._load_id_metrics[load_id][-1]
        # update finished at
        assert self._current_load_id is not None
        if finished:
            metrics["finished_at"] = ensure_pendulum_datetime_utc(precise_time())
            self._current_load_id = None
            self._current_load_started = None

    def _step_info_metrics(self, load_id: str) -> List[TStepMetrics]:
        return self._load_id_metrics[load_id]

    @property
    def current_load_id(self) -> str:
        """Returns currently processing load id"""
        return self._current_load_id

    @abstractmethod
    def get_step_info(
        self,
        pipeline: "SupportsPipeline",
    ) -> TStepInfo:
        """Returns and instance of StepInfo with metrics and package infos"""
        pass


class TLastRunContext(TypedDict, total=False):
    """Stores context from the last successful pipeline run or sync"""

    run_dir: str
    """Defines the context working directory, defaults to cwd()"""
    uri: str
    """Uniquely identifies the context"""
    local_dir: str
    """Defines data dir where local relative dirs and files are created, defaults to run_dir"""
    settings_dir: str
    """Defines where the current settings (secrets and configs) are located"""


class TPipelineLocalState(TypedDict, total=False):
    first_run: bool
    """Indicates a first run of the pipeline, where run ends with successful loading of data"""
    _last_extracted_at: datetime.datetime
    """Timestamp indicating when the state was synced with the destination."""
    _last_extracted_hash: str
    """Hash of state that was recently synced with destination"""
    initial_cwd: str
    """Run dir when pipeline was instantiated for a first time, defaults to cwd on OSS run context"""
    last_run_context: Optional[TLastRunContext]
    """Context from the last successful pipeline run or sync"""
    _dev_mode: bool
    """Tracks whether pipeline was created in dev_mode"""


class TPipelineState(TVersionedState, total=False):
    """Schema for a pipeline state that is stored within the pipeline working directory"""

    pipeline_name: str
    dataset_name: str
    default_schema_name: Optional[str]
    """Name of the first schema added to the pipeline to which all the resources without schemas will be added"""
    schema_names: Optional[List[str]]
    """All the schemas present within the pipeline working directory"""
    destination_name: Optional[str]
    destination_type: Optional[str]
    staging_name: Optional[str]
    staging_type: Optional[str]

    # properties starting with _ are not automatically applied to pipeline object when state is restored
    _local: TPipelineLocalState
    """A section of state that is not synchronized with the destination and does not participate in change merging and version control"""

    sources: NotRequired[Dict[str, Dict[str, Any]]]


class TSourceState(TPipelineState):
    sources: Dict[str, Dict[str, Any]]  # type: ignore[misc]


class SupportsPipeline(Protocol):
    """A protocol with core pipeline operations that lets high level abstractions ie. sources to access pipeline methods and properties"""

    pipeline_name: str
    """Name of the pipeline"""
    default_schema_name: str
    """Name of the default schema"""
    destination: AnyDestination
    """The destination reference which is ModuleType. `destination.__name__` returns the name string"""
    dataset_name: str
    """Name of the dataset to which pipeline will be loaded to"""
    run_context: RunContextBase
    """A run context associated with the pipeline when instance was created"""
    working_dir: str
    """A working directory of the pipeline"""
    pipeline_salt: str
    """A configurable pipeline secret to be used as a salt or a seed for encryption key"""
    first_run: bool
    """Indicates a first run of the pipeline, where run ends with successful loading of the data"""
    last_run_context: TLastRunContext
    """Stores last "good" run context, where run ends with successful loading of the data"""
    collector: Collector
    """A collector that tracks the progress of the pipeline"""

    @property
    def has_pending_data(self) -> bool:
        """ "Tells if pipeline contains any pending packages"""

    @property
    def state(self) -> TPipelineState:
        """Returns dictionary with current pipeline state

        Returns:
            TPipelineState: The current pipeline state
        """

    @property
    def schemas(self) -> Mapping[str, Schema]:
        """Returns all known schemas of the pipeline

        Returns:
            Mapping[str, Schema]: A mapping of schema names to their corresponding Schema objects
        """

    def get_local_state_val(self, key: str) -> Any:
        """Gets value from local state. Local state is not synchronized with destination.

        Args:
            key (str): The key to get the value from

        Returns:
            Any: The value from the local state
        """

    def destination_client(self, schema_name: str = None) -> JobClientBase:
        """Get the destination job client for the configured destination

        Args:
            schema_name (str, optional): The name of the schema to get the client for. Defaults to None.

        Returns:
            JobClientBase: The destination job client
        """

    def run(
        self,
        data: Any = None,
        *,
        destination: TDestinationReferenceArg = None,
        dataset_name: str = None,
        credentials: Any = None,
        table_name: str = None,
        write_disposition: TWriteDispositionConfig = None,
        columns: Sequence[TColumnSchema] = None,
        primary_key: TColumnNames = None,
        schema: Schema = None,
        loader_file_format: TLoaderFileFormat = None,
        schema_contract: TSchemaContract = None,
    ) -> LoadInfo:
        """Loads the data into the destination

        Args:
            data (Any): The data to load. Can be a DltSource, DltResource or any iterable of data items.
            destination (TDestinationReferenceArg, optional): The destination to load the data to.
            dataset_name (str, optional): The name of the dataset to load the data to.
            credentials (Any, optional): The credentials to use to load the data.
            table_name (str, optional): The name of the table to load the data to.
            write_disposition (TWriteDispositionConfig, optional): The write disposition to use to load the data.
            columns (Sequence[TColumnSchema], optional): A list of column hints.
            primary_key (TColumnNames, optional): A list of column names to be used as primary key.
            schema (Schema, optional): A schema to use to load the data. Defaults to the schema configured in the pipeline.
            loader_file_format (TLoaderFileFormat, optional): The file format to use to load the data. Defaults to preferred loader file format of the destination.
            schema_contract (TSchemaContract, optional): The schema contract to use to load the data.

        Returns:
            LoadInfo: Information on the load operation
        """

    def _set_context(self, is_active: bool) -> None:
        """Called when pipeline context activated or deactivate"""


class SupportsPipelineRun(Protocol):
    def __call__(
        self,
        *,
        destination: TDestinationReferenceArg = None,
        dataset_name: str = None,
        credentials: Any = None,
        table_name: str = None,
        write_disposition: TWriteDispositionConfig = None,
        columns: Sequence[TColumnSchema] = None,
        schema: Schema = None,
        loader_file_format: TLoaderFileFormat = None,
        schema_contract: TSchemaContract = None,
    ) -> LoadInfo: ...


@configspec
class PipelineContext(ContainerInjectableContext):
    _PIPELINE_FACTORY: ClassVar[Callable[[], SupportsPipeline]] = None
    """Maker of default pipeline, set at runtime"""
    _pipeline: SupportsPipeline = dataclasses.field(
        default=None, init=False, repr=False, compare=False
    )
    """Active pipeline"""
    _pipelines: Optional[List[SupportsPipeline]] = dataclasses.field(
        default=None, init=False, repr=False, compare=False
    )
    """History of previously activated pipelines, disabled by default"""
    enable_activation_history: bool = False
    """When True, references to activated pipelines will be also stored"""

    can_create_default: ClassVar[bool] = True

    def pipeline(self) -> SupportsPipeline:
        """Creates or returns an active pipeline"""
        if not self._pipeline:
            # delayed pipeline creation
            assert PipelineContext._PIPELINE_FACTORY is not None, (
                "Deferred pipeline creation function not provided to PipelineContext. Are you"
                " calling dlt.pipeline() from another thread?"
            )
            self.activate(PipelineContext._PIPELINE_FACTORY())
        return self._pipeline

    def activate(self, pipeline: SupportsPipeline) -> None:
        """Activates `pipeline` and deactivates active one."""
        from dlt.common.runtime.run_context import active

        # TODO: (1) compare run_context in pipeline with currently active context (via uri) and warn if they differ
        if pipeline.run_context.uri != active().uri:
            warnings.warn(
                f"Runtime context changed from `{pipeline.run_context.uri}` to `{active().uri}`"
                f" when activating pipeline `{pipeline.pipeline_name}`. Pipeline will keep its"
                " working and local dirs. Other behaviors are undefined. Recreate pipeline"
                " instance after run context change."
            )
        # do not activate currently active pipeline
        if pipeline == self._pipeline:
            return
        self.deactivate()
        # TODO: (2) activate the right pipeline context. that requires that we should change pluggable run context
        #       to thread-affine or even contextvar that works in async pools
        pipeline._set_context(True)
        self._pipeline = pipeline
        # store activated pipelines
        # NOTE: this prevents them from being garbage collected
        # NOTE: this context has thread affinity so each thread has own history
        if self.enable_activation_history:
            if self._pipelines is None:
                self._pipelines = []
            if pipeline not in self._pipelines:
                self._pipelines.append(pipeline)

    def is_active(self) -> bool:
        return self._pipeline is not None

    def deactivate(self) -> None:
        if self.is_active():
            self._pipeline._set_context(False)
        self._pipeline = None

    def clear_activation_history(self) -> None:
        self._pipelines = []

    def activation_history(self) -> List[SupportsPipeline]:
        """Get list of pipelines that were activated"""
        if not self.enable_activation_history:
            raise RuntimeError("Activation history is not enabled")
        return self._pipelines or []

    @classmethod
    def cls__init__(self, deferred_pipeline: Callable[..., SupportsPipeline] = None) -> None:
        """Initialize the context with a function returning the Pipeline object to allow creation on first use"""
        self._PIPELINE_FACTORY = deferred_pipeline


def current_pipeline() -> SupportsPipeline:
    """Gets active pipeline context or None if not found"""
    proxy = Container()[PipelineContext]
    if not proxy.is_active():
        return None
    return proxy.pipeline()


@configspec
class StateInjectableContext(ContainerInjectableContext):
    state: TPipelineState = None

    can_create_default: ClassVar[bool] = False


def pipeline_state(
    container: Container, initial_default: TPipelineState = None
) -> Tuple[TPipelineState, bool]:
    """Gets value of the state from context or active pipeline, if none found returns `initial_default`

    Injected state is called "writable": it is injected by the `Pipeline` class and all the changes will be persisted.
    The state coming from pipeline context or `initial_default` is called "read only" and all the changes to it will be discarded

    Returns tuple (state, writable)
    """
    try:
        # get injected state if present. injected state is typically "managed" so changes will be persisted
        return container[StateInjectableContext].state, True
    except ContextDefaultCannotBeCreated:
        # check if there's pipeline context
        proxy = container[PipelineContext]
        if not proxy.is_active():
            return initial_default, False
        else:
            # get unmanaged state that is read only
            # TODO: make sure that state if up to date by syncing the pipeline earlier
            return proxy.pipeline().state, False


def get_dlt_pipelines_dir() -> str:
    """Gets default directory where pipelines' data will be stored
    1. in user or workspace home directory ie. ~/.dlt/pipelines/
    2. if current user is root in /var/dlt/pipelines
    3. if current user does not have a home directory in /tmp/dlt/pipelines
    """
    from dlt.common.runtime import run_context

    return run_context.active().get_data_entity("pipelines")


def get_dlt_repos_dir() -> str:
    """Gets default directory where command repositories will be stored
    which is $global_dir/repos
    """
    from dlt.common.runtime import run_context

    return os.path.join(run_context.active().global_dir, "repos")



# from __future__ import annotations

import dataclasses
from typing import ClassVar, Union

from dlt.common.configuration import configspec
from dlt.common.configuration.exceptions import ConfigFieldMissingException
from dlt.common.configuration.specs.connection_string_credentials import ConnectionStringCredentials
from dlt.common.destination.capabilities import DestinationCapabilitiesContext
from dlt.common.destination.client import DestinationClientDwhWithStagingConfiguration
from dlt.common.storages.configuration import (
    FilesystemConfiguration,
    FilesystemConfigurationWithLocalFiles,
    WithLocalFiles,
)
from dlt.common.utils import digest128
from dlt.destinations.impl.duckdb.configuration import DuckDbConnectionPool, DuckDbBaseCredentials
from dlt.destinations.impl.duckdb.factory import _set_duckdb_raw_capabilities


DEFAULT_DUCKLAKE_NAME = "ducklake"
DUCKLAKE_STORAGE_PATTERN = "%s.files"


def _get_ducklake_capabilities() -> DestinationCapabilitiesContext:
    caps = DestinationCapabilitiesContext()
    caps = _set_duckdb_raw_capabilities(caps)
    # load with parquet by default
    caps.preferred_loader_file_format = "parquet"
    # duckdb and sqllite will crash when loading in parallel, see adjust_capabilities in factory
    caps.loader_parallelism_strategy = "sequential"
    return caps


@configspec(init=False)
class DuckLakeCredentials(DuckDbBaseCredentials):
    ducklake_name: str = DEFAULT_DUCKLAKE_NAME
    catalog: ConnectionStringCredentials = None
    # NOTE: consider moving to DuckLakeClientConfiguration so bucket_url is not a secret
    storage: FilesystemConfiguration = None

    __config_gen_annotations__: ClassVar[list[str]] = ["ducklake_name", "catalog", "storage"]

    def __init__(
        self,
        ducklake_name: str = DEFAULT_DUCKLAKE_NAME,
        catalog: Union[str, ConnectionStringCredentials] = None,
        storage: Union[str, FilesystemConfiguration] = None,
    ) -> None:
        """Initialize DuckLake credentials by passing ducklake name, catalog and storage
        configuration.

        Args:
            ducklake_name: str
                This value is mainly used as ATTACH name for the ducklake database and
                as names for catalog and storage files if not configured explicitly.
                If omitted, ducklake name is derived from destination name or pipeline name.
            catalog: Either a connection string (for example,
                "sqlite:///catalog.sqlite", "duckdb:///catalog.duckdb",
                or "postgres://loader:loader@localhost:5432/dlt_data") or a
                ConnectionStringCredentials instance. If omitted,
                will default the catalog to a local sqlite database whose filename is
                derived from the name_or_conn_str argument.
            storage: Either a storage URL string (for example,
                "file://...", "s3://bucket/prefix") or a FilesystemConfiguration
                instance. If omitted, it will create a folder for data with name
                derived from the name_or_conn_str argument.

        """
        self.ducklake_name = ducklake_name
        if isinstance(catalog, str):
            catalog = ConnectionStringCredentials(catalog)
        self.catalog = catalog
        if isinstance(storage, str):
            storage = FilesystemConfigurationWithLocalFiles(bucket_url=storage)
        self.storage = storage

    def _conn_str(self) -> str:
        return ":memory:"

    def on_partial(self) -> None:
        # this works only if wired to right exception type
        config_exception = self.__exception__
        if not isinstance(config_exception, ConfigFieldMissingException):
            return
        # set default catalog only if not present in config, partially resolved should generate exception
        if self.catalog is None and not config_exception.was_partially_resolved("catalog"):
            # use sqllite as default catalog
            self.catalog = ConnectionStringCredentials(
                {"drivername": "sqlite", "database": self.ducklake_name + ".sqlite"}
            ).resolve()
            config_exception.drop_traces_for_field("catalog")

        if self.storage is None and "bucket_url" in config_exception.traces["storage"][0].traces:  # type: ignore
            self.storage = FilesystemConfigurationWithLocalFiles(
                bucket_url=DUCKLAKE_STORAGE_PATTERN % self.ducklake_name, local_dir="."
            ).resolve()

        if not self.is_partial():
            self.resolve()

    def on_resolved(self) -> None:
        if self.extensions:
            self.extensions = list(set([*self.extensions, "ducklake"]))
        else:
            self.extensions = ["ducklake"]
        # set connection pool so it always opens a new connection on borrow.
        # connection duplication for parallelism does not work for ducklake.
        self.conn_pool = DuckDbConnectionPool(self, always_open_connection=True)

    @property
    def storage_url(self) -> str:
        """Convert file:// url into native os path so duckdb can read it"""
        if self.storage.is_local_filesystem:
            return self.storage.make_local_path(self.storage.bucket_url)
        else:
            return self.storage.bucket_url


# TODO add connection to a specific snapshot
@configspec
class DuckLakeClientConfiguration(WithLocalFiles, DestinationClientDwhWithStagingConfiguration):
    destination_type: str = dataclasses.field(
        default="ducklake",
        init=False,
        repr=False,
        compare=False,
    )
    credentials: DuckLakeCredentials = None
    create_indexes: bool = False  # does nothing but required
    override_data_path: bool = False

    def fingerprint(self) -> str:
        """Use fingerprint of underlying storage. This is precise to bucket level"""
        if self.credentials.storage is None:
            return ""
        return self.credentials.storage.fingerprint()

    def on_resolved(self) -> None:
        # redirect local catalog database file to `local_dir`
        if self.credentials.catalog.drivername in ("duckdb", "sqlite"):
            # name is <pipeline|dest name>.<duckdb|sqlite>
            local_db = self.make_location(
                self.credentials.catalog.database
                or self.credentials.ducklake_name + "." + self.credentials.catalog.drivername,
                "%s",
            )
            self.credentials.catalog.database = local_db

        # redirect storage to local filesystem
        if isinstance(self.credentials.storage, WithLocalFiles):
            self.credentials.storage.attach_from(self)
            if not self.credentials.storage.is_resolved():
                self.credentials.storage.resolve()
            else:
                self.credentials.storage.normalize_bucket_url()

    def __str__(self) -> str:
        """Return ducklake displayable location that contains catalog and storage locations"""
        if not self.credentials or not self.credentials.catalog or not self.credentials.storage:
            return ""
        return f"{self.credentials.ducklake_name}@{self.credentials.catalog}@{self.credentials.storage}"



"""Build a mermaid graph representation using raw strings without additional dependencies"""
from enum import Enum

from dlt.common.schema.typing import (
    TColumnSchema,
    TReferenceCardinality,
    TStoredSchema,
    TTableReferenceStandalone,
    TTableSchema,
)
from dlt.common.schema.utils import is_complete_column


INDENT = "    "


def schema_to_mermaid(
    schema: TStoredSchema,
    *,
    references: list[TTableReferenceStandalone],
    hide_columns: bool = False,
    hide_descriptions: bool = False,
    include_dlt_tables: bool = True,
) -> str:
    mermaid_er_diagram = "erDiagram\n"

    for table_name, table_schema in schema["tables"].items():
        if not include_dlt_tables and table_name.startswith("_dlt"):
            continue
        # skip tables with no complete columns
        if not any(is_complete_column(c) for c in table_schema.get("columns", {}).values()):
            continue

        mermaid_er_diagram += INDENT + _to_mermaid_table(
            table_schema,
            hide_columns=hide_columns,
            hide_descriptions=hide_descriptions,
        )

    for ref in references:
        if not include_dlt_tables:
            if ref["table"].startswith("_dlt") or ref["referenced_table"].startswith("_dlt"):
                continue

        mermaid_er_diagram += INDENT + _to_mermaid_reference(ref)

    return mermaid_er_diagram


def _to_mermaid_table(
    table: TTableSchema, hide_columns: bool = False, hide_descriptions: bool = False
) -> str:
    mermaid_table: str = table["name"]
    mermaid_table += "{\n"

    if hide_columns is False:
        for column in table["columns"].values():
            if not is_complete_column(column):
                continue
            mermaid_table += INDENT + _to_mermaid_column(
                column,
                hide_descriptions=hide_descriptions,
            )

    mermaid_table += "}\n"
    return mermaid_table


# TODO add scale & precision to `data_type`
def _to_mermaid_column(column: TColumnSchema, hide_descriptions: bool = False) -> str:
    # NOTE: incomplete columns (no data_type) currently arise from "seen-null-first" propagation
    # (columns whose first observed values were all null). if that propagation is removed so that
    # incomplete columns no longer reach the schema, this fallback should be revisited.
    mermaid_col = column.get("data_type", "no_data_seen") + " " + column["name"]
    keys = []
    if column.get("primary_key"):
        keys.append("PK")

    if column.get("unique"):
        keys.append("UK")

    if keys:
        mermaid_col += " " + ",".join(keys)

    if hide_descriptions is False:
        if description := column.get("description"):
            mermaid_col += f' "{description}"'

    mermaid_col += "\n"
    return mermaid_col


class TMermaidArrows(str, Enum):
    ONE_TO_MANY = "||--|{"
    MANY_TO_ONE = "}|--||"
    ZERO_TO_MANY = "|o--|{"
    MANY_TO_ZERO = "}|--o|"
    ONE_TO_MORE = "||--o{"
    MORE_TO_ONE = "}o--||"
    ONE_TO_ONE = "||--||"
    MANY_TO_MANY = "}|--|{"
    ZERO_TO_ONE = "|o--o|"


_CARDINALITY_ARROW: dict[TReferenceCardinality, TMermaidArrows] = {
    "one_to_many": TMermaidArrows.ONE_TO_MANY,
    "many_to_one": TMermaidArrows.MANY_TO_ONE,
    "zero_to_many": TMermaidArrows.ZERO_TO_MANY,
    "many_to_zero": TMermaidArrows.MANY_TO_ZERO,
    "one_to_one": TMermaidArrows.ONE_TO_ONE,
    "many_to_many": TMermaidArrows.MANY_TO_MANY,
    "zero_to_one": TMermaidArrows.ZERO_TO_ONE,
    "one_to_zero": TMermaidArrows.ZERO_TO_ONE,
}


def _to_mermaid_reference(ref: TTableReferenceStandalone) -> str:
    """Builds references in the following format using cardinality and label to describe
    the relationship

    <left-entity> [<relationship> <right-entity> : <relationship-label>]
    """
    left_table = ref.get("table")
    right_table = ref.get("referenced_table")
    cardinality = ref.get("cardinality", "one_to_many")
    label = ref.get("label", "")
    arrow: str = _CARDINALITY_ARROW.get(cardinality).value

    mermaid_reference = f'{left_table} {arrow} {right_table} : "{label}"\n'
    return mermaid_reference



from typing import List, Dict, Set, Any

from dlt.common import logger
from dlt.common.data_writers.writers import ArrowToObjectAdapter
from dlt.common.json import json
from dlt.common.metrics import DataWriterMetrics
from dlt.common.normalizers.json.relational import DataItemNormalizer as RelationalNormalizer
from dlt.common.normalizers.utils import generate_dlt_ids
from dlt.common.schema.typing import C_DLT_ID
from dlt.common.schema.utils import dlt_id_column, normalize_table_identifiers
from dlt.common.schema import TSchemaUpdate, Schema
from dlt.common.storages.load_storage import LoadStorage
from dlt.common.storages import NormalizeStorage
from dlt.common.storages.data_item_storage import DataItemStorage
from dlt.common.storages.load_package import ParsedLoadJobFileName
from dlt.common.exceptions import MissingDependencyException

from dlt.normalize.configuration import NormalizeConfiguration
from dlt.normalize.items_normalizers.base import ItemsNormalizer

try:
    from dlt.common.libs import pyarrow
    from dlt.common.libs.pyarrow import pyarrow as pa
except MissingDependencyException:
    pyarrow = None
    pa = None


class ArrowItemsNormalizer(ItemsNormalizer):
    REWRITE_ROW_GROUPS = 1

    def __init__(
        self,
        item_storage: DataItemStorage,
        load_storage: LoadStorage,
        normalize_storage: NormalizeStorage,
        schema: Schema,
        load_id: str,
        config: NormalizeConfiguration,
    ) -> None:
        super().__init__(item_storage, load_storage, normalize_storage, schema, load_id, config)
        self._null_only_columns: Dict[str, Set[str]] = {}

    @property
    def null_only_columns(self) -> Dict[str, Set[str]]:
        return self._null_only_columns

    def _collect_null_columns_from_arrow_metadata(
        self, arrow_schema: Any, root_table_name: str
    ) -> None:
        """Read dlt.null_columns from arrow schema metadata, normalize names, add to tracker."""
        metadata = arrow_schema.metadata or {}
        null_cols_json = metadata.get(b"dlt.null_columns")
        if not null_cols_json:
            return
        null_col_names = json.loadb(null_cols_json)
        if not null_col_names:
            return
        normalized_names = set()
        for name in null_col_names:
            normalized_names.add(self.schema.naming.normalize_path(name))
        self._null_only_columns.setdefault(root_table_name, set()).update(normalized_names)

    def _write_with_dlt_columns(
        self,
        extracted_items_file: str,
        root_table_name: str,
        add_dlt_id: bool,
    ) -> List[TSchemaUpdate]:
        new_columns: List[Any] = []
        schema = self.schema
        load_id = self.load_id
        schema_update: TSchemaUpdate = {}
        data_normalizer = schema.data_item_normalizer

        if add_dlt_id and isinstance(data_normalizer, RelationalNormalizer):
            partial_table = normalize_table_identifiers(
                {
                    "name": root_table_name,
                    "columns": {C_DLT_ID: dlt_id_column()},
                },
                schema.naming,
            )
            schema.update_table(partial_table, normalize_identifiers=False)
            table_updates = schema_update.setdefault(root_table_name, [])
            table_updates.append(partial_table)
            # TODO: use get_root_row_id_type to get row id type and generate deterministic
            #  row ids as well (using pandas helper function prepared for scd2)
            #  we could also generate random columns with pandas or duckdb if present
            new_columns.append(
                (
                    -1,
                    pa.field(data_normalizer.c_dlt_id, pyarrow.pyarrow.string(), nullable=False),
                    lambda batch: pa.array(generate_dlt_ids(batch.num_rows)),
                )
            )

        items_count = 0
        columns_schema = schema.get_table_columns(root_table_name)
        # if we use adapter to convert arrow to dicts, then normalization is not necessary
        is_native_arrow_writer = not issubclass(self.item_storage.writer_cls, ArrowToObjectAdapter)
        should_normalize: bool = None
        self._maybe_cancel()
        with self.normalize_storage.extracted_packages.storage.open_file(
            extracted_items_file, "rb"
        ) as f:
            for batch in pyarrow.pq_stream_with_new_columns(
                f, new_columns, row_groups_per_read=self.REWRITE_ROW_GROUPS
            ):
                self._maybe_cancel()
                items_count += batch.num_rows
                # we may need to normalize
                if is_native_arrow_writer and should_normalize is None:
                    should_normalize = pyarrow.should_normalize_arrow_schema(
                        batch.schema, columns_schema, schema.naming
                    )[0]
                    if should_normalize:
                        logger.info(
                            f"When writing arrow table to {root_table_name} the schema requires"
                            " normalization because its shape does not match the actual schema of"
                            " destination table. Arrow table columns will be reordered and missing"
                            " columns will be added if needed."
                        )
                if should_normalize:
                    batch = pyarrow.normalize_py_arrow_item(
                        batch, columns_schema, schema.naming, self.config.destination_capabilities
                    )
                    # normalize may remove null columns and set dlt.null_columns metadata
                    self._collect_null_columns_from_arrow_metadata(batch.schema, root_table_name)
                self.item_storage.write_data_item(
                    load_id,
                    schema.name,
                    root_table_name,
                    batch,
                    columns_schema,
                )
        # TODO: better to check if anything is in the buffer and skip writing file
        if items_count == 0 and not is_native_arrow_writer:
            self.item_storage.write_empty_items_file(
                load_id,
                schema.name,
                root_table_name,
                columns_schema,
            )

        return [schema_update]

    def __call__(self, extracted_items_file: str, root_table_name: str) -> List[TSchemaUpdate]:
        self._maybe_cancel()
        # read schema and counts from file metadata
        from dlt.common.libs.pyarrow import get_parquet_metadata

        with self.normalize_storage.extracted_packages.storage.open_file(
            extracted_items_file, "rb"
        ) as f:
            num_rows, arrow_schema = get_parquet_metadata(f)
            file_metrics = DataWriterMetrics(extracted_items_file, num_rows, f.tell(), 0, 0)

        # collect null column metadata from arrow schema
        self._collect_null_columns_from_arrow_metadata(arrow_schema, root_table_name)

        add_dlt_id = self.config.parquet_normalizer.add_dlt_id
        # TODO: add dlt id only if not present in table
        # if we need to add any columns or the file format is not parquet, we can't just import files
        must_rewrite = add_dlt_id or self.item_storage.writer_spec.file_format != "parquet"
        if not must_rewrite:
            # in rare cases normalization may be needed
            must_rewrite = pyarrow.should_normalize_arrow_schema(
                arrow_schema, self.schema.get_table_columns(root_table_name), self.schema.naming
            )[0]
        if must_rewrite:
            logger.info(
                f"Table {root_table_name} parquet file {extracted_items_file} must be rewritten:"
                f" add_dlt_id: {add_dlt_id} destination file"
                f" format: {self.item_storage.writer_spec.file_format} or due to required"
                " normalization "
            )
            schema_update = self._write_with_dlt_columns(
                extracted_items_file, root_table_name, add_dlt_id
            )
            return schema_update

        logger.info(
            f"Table {root_table_name} parquet file {extracted_items_file} will be directly imported"
            " without normalization"
        )
        parts = ParsedLoadJobFileName.parse(extracted_items_file)
        self.item_storage.import_items_file(
            self.load_id,
            self.schema.name,
            parts.table_name,
            self.normalize_storage.extracted_packages.storage.make_full_path(extracted_items_file),
            file_metrics,
        )

        return []



from typing import ClassVar, Optional, Sequence, Tuple, List, Any
from urllib.parse import urlparse

from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.destination.client import (
    HasFollowupJobs,
    PreparedTableSchema,
    RunnableLoadJob,
    SupportsStagingDestination,
    FollowupJobRequest,
    LoadJob,
)
from dlt.common.schema import TColumnSchema, Schema
from dlt.common.schema.typing import TColumnType
from dlt.common.storages.file_storage import FileStorage
from dlt.destinations.exceptions import LoadJobTerminalException
from dlt.destinations.impl.dremio.configuration import DremioClientConfiguration
from dlt.destinations.impl.dremio.sql_client import DremioSqlClient
from dlt.destinations.job_client_impl import SqlJobClientWithStagingDataset
from dlt.destinations.job_impl import ReferenceFollowupJobRequest
from dlt.destinations.sql_jobs import SqlMergeFollowupJob
from dlt.destinations.sql_client import SqlClientBase
from dlt.destinations.utils import get_deterministic_temp_table_name


class DremioMergeJob(SqlMergeFollowupJob):
    @classmethod
    def _new_temp_table_name(cls, table_name: str, op: str, sql_client: SqlClientBase[Any]) -> str:
        with sql_client.with_staging_dataset():
            return sql_client.make_qualified_table_name(
                cls._shorten_table_name(
                    get_deterministic_temp_table_name(table_name, op), sql_client
                )
            )

    @classmethod
    def _to_temp_table(cls, select_sql: str, temp_table_name: str, unique_column: str) -> str:
        return (
            f"DROP TABLE IF EXISTS {temp_table_name}; CREATE TABLE {temp_table_name} AS"
            f" {select_sql}"
        )

    @classmethod
    def default_order_by(cls) -> str:
        return "NULL"


class DremioLoadJob(RunnableLoadJob, HasFollowupJobs):
    def __init__(
        self,
        file_path: str,
        stage_name: Optional[str] = None,
    ) -> None:
        super().__init__(file_path)
        self._stage_name = stage_name
        self._job_client: "DremioClient" = None

    def run(self) -> None:
        self._sql_client = self._job_client.sql_client

        qualified_table_name = self._sql_client.make_qualified_table_name(self.load_table_name)

        # extract and prepare some vars
        bucket_path = (
            ReferenceFollowupJobRequest.resolve_reference(self._file_path)
            if ReferenceFollowupJobRequest.is_reference_job(self._file_path)
            else ""
        )

        if not bucket_path:
            raise RuntimeError("Could not resolve `bucket_path`.")

        file_name = (
            FileStorage.get_file_name_from_file_path(bucket_path)
            if bucket_path
            else self._file_name
        )

        bucket_url = urlparse(bucket_path)
        bucket_scheme = bucket_url.scheme
        if bucket_scheme == "s3" and self._stage_name:
            from_clause = (
                f"FROM '@{self._stage_name}/{bucket_url.hostname}/{bucket_url.path.lstrip('/')}'"
            )
        else:
            raise LoadJobTerminalException(
                self._file_path, "Only s3 staging currently supported in Dremio destination"
            )

        source_format = file_name.split(".")[-1]

        self._sql_client.execute_sql(f"""COPY INTO {qualified_table_name}
            {from_clause}
            FILE_FORMAT '{source_format}'
            """)


class DremioClient(SqlJobClientWithStagingDataset, SupportsStagingDestination):
    def __init__(
        self,
        schema: Schema,
        config: DremioClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        dataset_name, staging_dataset_name = SqlJobClientWithStagingDataset.create_dataset_names(
            schema, config
        )
        sql_client = DremioSqlClient(
            dataset_name,
            staging_dataset_name,
            config.credentials,
            capabilities,
        )
        super().__init__(schema, config, sql_client)
        self.config: DremioClientConfiguration = config
        self.sql_client: DremioSqlClient = sql_client  # type: ignore
        self.type_mapper = self.capabilities.get_type_mapper()

    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        job = super().create_load_job(table, file_path, load_id, restore)

        if not job:
            job = DremioLoadJob(
                file_path=file_path,
                stage_name=self.config.staging_data_source,
            )
        return job

    def _get_table_update_sql(
        self, table_name: str, new_columns: Sequence[TColumnSchema], generate_alter: bool
    ) -> List[str]:
        sql = super()._get_table_update_sql(table_name, new_columns, generate_alter)

        if not generate_alter:
            partition_list = [
                self.sql_client.escape_column_name(c["name"])
                for c in new_columns
                if c.get("partition")
            ]
            if partition_list:
                sql[0] += "\nPARTITION BY (" + ",".join(partition_list) + ")"

            sort_list = [
                self.sql_client.escape_column_name(c["name"]) for c in new_columns if c.get("sort")
            ]
            if sort_list:
                sql[0] += "\nLOCALSORT BY (" + ",".join(sort_list) + ")"

        return sql

    def _from_db_type(
        self, bq_t: str, precision: Optional[int], scale: Optional[int]
    ) -> TColumnType:
        return self.type_mapper.from_destination_type(bq_t, precision, scale)

    def _create_merge_followup_jobs(
        self, table_chain: Sequence[PreparedTableSchema]
    ) -> List[FollowupJobRequest]:
        return [DremioMergeJob.from_table_chain(table_chain, self.sql_client)]

    def _make_add_column_sql(
        self, new_columns: Sequence[TColumnSchema], table: PreparedTableSchema = None
    ) -> List[str]:
        return [
            "ADD COLUMNS ("
            + ", ".join(self._get_column_def_sql(c, table) for c in new_columns)
            + ")"
        ]

    def should_truncate_table_before_load_on_staging_destination(self, table_name: str) -> bool:
        return self.config.truncate_tables_on_staging_destination_before_load



"""Implements SupportsTracking"""
from typing import Any, cast, List
from requests import Session

from dlt.common import logger
from dlt.common.json import json
from dlt.common.pipeline import LoadInfo
from dlt.common.managed_thread_pool import ManagedThreadPool
from dlt.common.schema.typing import TStoredSchema
from dlt.common.typing import TypedDict

from dlt.pipeline.trace import PipelineTrace, PipelineStepTrace, TPipelineStep, SupportsPipeline

_THREAD_POOL: ManagedThreadPool = None
TRACE_URL_SUFFIX = "/trace"
STATE_URL_SUFFIX = "/state"
requests: Session = None


class TPipelineSyncPayload(TypedDict):
    pipeline_name: str
    destination_name: str
    destination_displayable_credentials: str
    destination_fingerprint: str
    dataset_name: str
    schemas: List[TStoredSchema]


def init_platform_tracker() -> None:
    # lazily import requests to avoid binding config before initialization
    global requests
    from dlt.sources.helpers.requests import Client

    # fail fast, don't block user
    requests = Client(request_timeout=(2, 10), request_max_attempts=0)  # type: ignore[assignment]

    global _THREAD_POOL
    if _THREAD_POOL is None:
        _THREAD_POOL = ManagedThreadPool("platform_tracker", 1)
        # create thread pool in controlled way, not lazy
        _THREAD_POOL._create_thread_pool()


def disable_platform_tracker() -> None:
    global _THREAD_POOL
    if _THREAD_POOL:
        _THREAD_POOL.stop()
    _THREAD_POOL = None


def _send_trace_to_platform(trace: PipelineTrace, pipeline: SupportsPipeline) -> None:
    """
    Send the full trace after a run operation to the platform
    TODO: Migrate this to open telemetry in the next iteration
    """
    if not pipeline.run_context.runtime_config.dlthub_dsn:
        return

    def _future_send() -> None:
        try:
            trace_dump = json.dumps(trace.asdict())
            url = pipeline.run_context.runtime_config.dlthub_dsn + TRACE_URL_SUFFIX
            response = requests.put(url, data=trace_dump)
            if response.status_code != 200:
                logger.debug(
                    f"Failed to send trace to platform, response code: {response.status_code}"
                )
        except Exception as e:
            logger.debug(f"Exception while sending trace to platform: {e}")

    _THREAD_POOL.thread_pool.submit(_future_send)

    # trace_dump = json.dumps(trace.asdict(), pretty=True)
    # with open(f"trace.json", "w") as f:
    #     f.write(trace_dump)


def _sync_schemas_to_platform(trace: PipelineTrace, pipeline: SupportsPipeline) -> None:
    if not pipeline.run_context.runtime_config.dlthub_dsn:
        return

    # sync only if load step was processed
    load_info: LoadInfo = None
    for step in trace.steps:
        if step.step == "load":
            load_info = cast(LoadInfo, step.step_info)

    if not load_info:
        return

    payload = TPipelineSyncPayload(
        pipeline_name=pipeline.pipeline_name,
        destination_name=load_info.destination_name,
        destination_displayable_credentials=load_info.destination_displayable_credentials,
        destination_fingerprint=load_info.destination_fingerprint,
        dataset_name=load_info.dataset_name,
        schemas=[],
    )

    # attach all schemas
    for schema_name in pipeline.schemas:
        schema = pipeline.schemas[schema_name]
        payload["schemas"].append(schema.to_dict())

    def _future_send() -> None:
        try:
            url = pipeline.run_context.runtime_config.dlthub_dsn + STATE_URL_SUFFIX
            response = requests.put(url, data=json.dumps(payload))
            if response.status_code != 200:
                logger.debug(
                    f"Failed to send state to platform, response code: {response.status_code}"
                )
        except Exception as e:
            logger.debug(f"Exception while sending state to platform: {e}")

    _THREAD_POOL.thread_pool.submit(_future_send)


def on_start_trace(trace: PipelineTrace, step: TPipelineStep, pipeline: SupportsPipeline) -> None:
    pass


def on_start_trace_step(
    trace: PipelineTrace, step: TPipelineStep, pipeline: SupportsPipeline
) -> None:
    pass


def on_end_trace_step(
    trace: PipelineTrace,
    step: PipelineStepTrace,
    pipeline: SupportsPipeline,
    step_info: Any,
    send_state: bool,
) -> None:
    if send_state:
        # also sync schemas to dlthub
        _sync_schemas_to_platform(trace, pipeline)


def on_end_trace(trace: PipelineTrace, pipeline: SupportsPipeline, send_state: bool) -> None:
    _send_trace_to_platform(trace, pipeline)



from typing import Dict, Iterable, Optional

from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.schema import TColumnHint, Schema
from dlt.common.destination.client import (
    PreparedTableSchema,
    RunnableLoadJob,
    HasFollowupJobs,
    LoadJob,
)
from dlt.common.schema.typing import TColumnSchema, TColumnType, TTableFormat
from dlt.common.schema.utils import has_default_column_prop_value
from dlt.common.storages.file_storage import FileStorage
from dlt.common.storages.load_storage import ParsedLoadJobFileName

from dlt.destinations.insert_job_client import InsertValuesJobClient

from dlt.destinations.impl.duckdb.sql_client import DuckDbSqlClient
from dlt.destinations.impl.duckdb.configuration import DuckDbClientConfiguration


HINT_TO_POSTGRES_ATTR: Dict[TColumnHint, str] = {"unique": "UNIQUE"}


class DuckDbCopyJob(RunnableLoadJob, HasFollowupJobs):
    def __init__(self, file_path: str) -> None:
        super().__init__(file_path)
        self._job_client: "DuckDbClient" = None

    def run(self) -> None:
        self._sql_client = self._job_client.sql_client

        qualified_table_name = self._sql_client.make_qualified_table_name(self.load_table_name)

        parsed_file = ParsedLoadJobFileName.parse(self._file_path)
        if parsed_file.file_format == "parquet":
            source_format = "read_parquet"
            options = ", union_by_name=true"
        elif parsed_file.file_format in ["jsonl", "typed-jsonl"]:
            # NOTE: loading JSON does not work in practice on duckdb: the missing keys fail the load instead of being interpreted as NULL
            source_format = "read_json"  # newline delimited, compression auto
            options = ""
        else:
            raise ValueError(self._file_path)

        self._sql_client.execute_sql(
            f"INSERT INTO {qualified_table_name} BY NAME SELECT * FROM"
            f" {source_format}('{self._file_path}' {options})"
        )


class DuckDbClient(InsertValuesJobClient):
    def __init__(
        self,
        schema: Schema,
        config: DuckDbClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        dataset_name, staging_dataset_name = InsertValuesJobClient.create_dataset_names(
            schema, config
        )
        sql_client = DuckDbSqlClient(
            dataset_name,
            staging_dataset_name,
            config.credentials,
            capabilities,
        )
        super().__init__(schema, config, sql_client)
        self.config: DuckDbClientConfiguration = config
        self.sql_client: DuckDbSqlClient = sql_client  # type: ignore
        self.active_hints = HINT_TO_POSTGRES_ATTR if self.config.create_indexes else {}
        self.type_mapper = self.capabilities.get_type_mapper()

    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        job = super().create_load_job(table, file_path, load_id, restore)
        if not job:
            job = DuckDbCopyJob(file_path)
        return job

    def initialize_storage(self, truncate_tables: Iterable[str] = None) -> None:
        self.sql_client.warn_if_catalog_equals_dataset_name()
        super().initialize_storage(truncate_tables)

    def _from_db_type(
        self, pq_t: str, precision: Optional[int], scale: Optional[int]
    ) -> TColumnType:
        return self.type_mapper.from_destination_type(pq_t, precision, scale)

    # def _make_add_column_sql(
    #     self, new_columns: Sequence[TColumnSchema], table: PreparedTableSchema = None
    # ) -> List[str]:
    #     # skip nullability on duckdb
    #     new_columns = list(new_columns)
    #     for idx, c in enumerate(new_columns):
    #         if not has_default_column_prop_value("nullable", c.get("nullable")):
    #             logger.warning(
    #                 f"Adding new NOT NULL column '{c['name']}' to existing table '{table['name']}'"
    #                 " on duckdb. NOT NULL will be ignored."
    #             )
    #             # make copy so original new column is not changed
    #             new_columns[idx] = copy(c)
    #             new_columns[idx]["nullable"] = True

    #     return [f"ADD COLUMN {self._get_column_def_sql(c, table)}" for c in new_columns]



import logging
from typing import Optional, Tuple, cast

import sqlglot.expressions as sge

from sqlglot.errors import OptimizeError
from sqlglot.schema import Schema as SQLGlotSchema, ensure_schema
from sqlglot.optimizer.annotate_types import annotate_types
from sqlglot.optimizer.qualify import qualify

import dlt
from dlt.common import logger
from dlt.common.libs.sqlglot import (
    to_sqlglot_type,
    from_sqlglot_type,
    set_metadata,
    get_metadata,
    TSqlGlotDialect,
)
from dlt.common.schema.typing import (
    TTableSchemaColumns,
    TColumnSchema,
)

from dlt.dataset.exceptions import LineageFailedException


def create_sqlglot_schema(
    schema: dlt.Schema,
    dataset_name: str,
    dialect: TSqlGlotDialect,
) -> SQLGlotSchema:
    """Create an SQLGlot schema using a dlt Schema and the destination capabilities.

    The SQLGlot schema automatically scopes the tables to the `dataset_name`.
    This can allow cross-dataset transformations on the same physical location.
    No name translation nor case folding is performing. All identifiers correspond
    to identifiers in dlt schema.
    """

    sqlglot_schema = {}  # MappingSchema(empty_schema, normalize=False)

    for table_name in schema.tables.keys():
        column_mapping = {}
        # skip not materialized columns
        for column_name, column in schema.get_table_columns(
            table_name, include_incomplete=False
        ).items():
            sqlglot_type = to_sqlglot_type(
                dlt_type=column["data_type"],
                nullable=column.get("nullable"),
                precision=column.get("precision"),
                scale=column.get("scale"),
                timezone=column.get("timezone"),
            )
            sqlglot_type = set_metadata(sqlglot_type, column)
            # column_name = sql_client.capabilities.casefold_identifier(column_name)
            column_mapping[column_name] = sqlglot_type
        # skip tables without columns
        if column_mapping:
            # table_name = sql_client.make_qualified_table_name_path(table_name, quote=False, casefold=False)[-1]
            sqlglot_schema[table_name] = column_mapping

    # ensure proper nesting with db and catalog
    nested_schema = {dataset_name: sqlglot_schema}

    return ensure_schema(nested_schema, dialect=dialect, normalize=False)


# NOTE even if `infer_sqlglot_schema=True`, some queries can have undetermined final columns
# TODO refactor such that it doesn't return a tuple
def compute_columns_schema(
    expression: sge.Expression,
    sqlglot_schema: SQLGlotSchema,
    dialect: TSqlGlotDialect,
    infer_sqlglot_schema: bool = True,
    allow_anonymous_columns: bool = True,
    allow_partial: bool = True,
) -> Tuple[TTableSchemaColumns, Optional[sge.Query]]:
    """Compute the expected dlt columns schema for the output of an SQL SELECT query. No case-folding or
    quoting is performed on the query.

    Args:
        infer_sqlglot_schema (bool): If False, all columns and tables referenced must be derived from the SQLGlot schema.
            If True, allow columns and tables not found in SQLGlot schema
        allow_anonymous_columns (bool): If False, all columns in final selection must have an explicit name or alias.
            If True, the name of columns from the final selection can be generated by the dialect
        allow_partial (bool): If False, raise exceptions if the schema returned is incomplete.
            If True, this function always returns a dictionary, even in cases of
            SQL parsing errors, missing table reference, unresolved `SELECT *`, etc.

    Returns: tuple of dlt columns schema and qualified `sql_query`

    """
    # make sure we don't modify the original expression
    expression = expression.copy()

    # the only thing we care is a list of selects
    if not isinstance(expression, sge.Query):
        if allow_partial:
            logger.debug(
                "Parsed SQL query is not a SQL query. Returning empty table schema because"
                " `allow_fail=True`"
            )
            return {}, None

        raise LineageFailedException(
            "Parsed SQL query is not a SELECT statement. Received SQL expression of type"
            f" {expression.type}.",
        )
    else:
        select_expression = expression

    # prevent normalization
    select_expression.meta["case_sensitive"] = True

    # false-y values `schema={}` or `schema=None` are identical to `infer_schema=True`
    try:
        select_expression = cast(
            sge.Query,
            qualify(
                select_expression,
                schema=sqlglot_schema,
                dialect=dialect,
                infer_schema=infer_sqlglot_schema,
                quote_identifiers=False,
                expand_stars=True,
            ),
        )
    except OptimizeError as e:
        raise LineageFailedException(
            f"Failed to resolve SQL query against the schema received: {e}"
        ) from e

    expression = annotate_types(expression, schema=sqlglot_schema)

    # NOTE: this has to be fixed
    if allow_anonymous_columns is False:
        for col in select_expression.selects:
            if col.alias_or_name == "" and False:
                raise LineageFailedException(
                    "Found anonymous column in SELECT statement. Use"
                    f" `allow_anonymous_columns=True` for permissive handling.\nColumn:\n\t{col}",
                )

    dlt_table_schema: dict[str, TColumnSchema] = {}
    for col in select_expression.selects:
        if col.output_name == "*":
            if allow_partial is True:
                logger.debug(
                    "SELECT statement includes a `*` selection that can't be resolved. "
                    "Returning empty table schema because `allow_fail=True`"
                )
                continue

            raise LineageFailedException(
                "SELECT statement includes a `*` selection that can't be resolved. Modify the"
                " query to select columns explicitly or limit `*` to known tables (e.g., `SELECT"
                " known_table.*`). Use `allow_fail=True` to return a partial dlt schema with"
                f" resolvable column hints.\nColumn:\n\t{col}",
            )

        data_type_hints = from_sqlglot_type(sqlglot_type=col.type)
        additional_hints = get_metadata(sqlglot_type=col.type)
        # get original name for queries that aliased the source dlt column
        propagated_name = additional_hints.pop("name", None)
        # NOTE dictionary unpacking order matters; unpacking `data_type_hints` last ensures precedence.
        dlt_table_schema[col.output_name] = {
            "name": col.output_name,
            **additional_hints,
            **data_type_hints,
        }
        if propagated_name and col.output_name != propagated_name:
            dlt_table_schema[col.output_name]["x-original-name"] = propagated_name  # type: ignore[typeddict-unknown-key]

    return dlt_table_schema, expression



from dlt.common.exceptions import MissingDependencyException
from dlt import version

try:
    from sqlalchemy import MetaData, Table, Column, create_engine, TypeDecorator  # noqa: I251
    from sqlalchemy.engine import Engine, URL, make_url, Row  # noqa: I251
    from sqlalchemy.sql import sqltypes, Select, Executable  # noqa: I251
    from sqlalchemy.sql.elements import TextClause  # noqa: I251
    from sqlalchemy.sql.sqltypes import TypeEngine  # noqa: I251
    from sqlalchemy.exc import CompileError  # noqa: I251
    from sqlalchemy.dialects.oracle import NUMBER as ORACLE_NUMBER  # noqa: I251
    from sqlalchemy.dialects.oracle.base import OracleDialect  # noqa: I251
    import sqlalchemy as sa  # noqa: I251
except ModuleNotFoundError:
    raise MissingDependencyException(
        "dlt sql_database helpers ",
        [f"{version.DLT_PKG_NAME}[sql_database]"],
        "Install the `sql_database` helpers for loading from `sql_database` sources. Note that you"
        " may need to install additional SQLAlchemy dialects for your source database.",
    )

# TODO: maybe use sa.__version__?
IS_SQL_ALCHEMY_20 = hasattr(sa, "Double")

__all__ = [
    "IS_SQL_ALCHEMY_20",
    "MetaData",
    "Table",
    "Column",
    "create_engine",
    "TypeDecorator",
    "Engine",
    "URL",
    "make_url",
    "Row",
    "sqltypes",
    "Select",
    "Executable",
    "TextClause",
    "TypeEngine",
    "CompileError",
    "ORACLE_NUMBER",
    "OracleDialect",
    "sa",
]



import os
from typing import Sequence, List, Dict, Any, Optional, cast, Union
from copy import deepcopy
from textwrap import dedent
from urllib.parse import urlparse, urlunparse

from dlt.common.destination import DestinationCapabilitiesContext
from dlt.common.destination.client import (
    PreparedTableSchema,
    SupportsStagingDestination,
    FollowupJobRequest,
    LoadJob,
)

from dlt.common.schema import TColumnSchema, Schema, TColumnHint
from dlt.common.schema.utils import (
    get_inherited_table_hint,
)

from dlt.common.configuration.exceptions import ConfigurationException
from dlt.common.configuration.specs import (
    AzureCredentialsWithoutDefaults,
    AzureServicePrincipalCredentialsWithoutDefaults,
)

from dlt.destinations.job_impl import ReferenceFollowupJobRequest
from dlt.destinations.job_client_impl import (
    SqlJobClientBase,
    CopyRemoteFileLoadJob,
)

from dlt.destinations.impl.mssql.mssql import (
    MsSqlJobClient,
    VARCHAR_MAX_N,
    VARBINARY_MAX_N,
)

from dlt.destinations.impl.synapse.sql_client import SynapseSqlClient
from dlt.destinations.impl.synapse.configuration import SynapseClientConfiguration
from dlt.destinations.impl.synapse.synapse_adapter import (
    TABLE_INDEX_TYPE_HINT,
    TTableIndexType,
)


HINT_TO_SYNAPSE_ATTR: Dict[TColumnHint, str] = {
    "primary_key": "PRIMARY KEY NONCLUSTERED NOT ENFORCED",
    "unique": "UNIQUE NOT ENFORCED",
}
TABLE_INDEX_TYPE_TO_SYNAPSE_ATTR: Dict[TTableIndexType, str] = {
    "heap": "HEAP",
    "clustered_columnstore_index": "CLUSTERED COLUMNSTORE INDEX",
}


class SynapseClient(MsSqlJobClient, SupportsStagingDestination):
    def __init__(
        self,
        schema: Schema,
        config: SynapseClientConfiguration,
        capabilities: DestinationCapabilitiesContext,
    ) -> None:
        super().__init__(schema, config, capabilities)
        self.config: SynapseClientConfiguration = config
        dataset_name, staging_dataset_name = MsSqlJobClient.create_dataset_names(schema, config)
        self.sql_client = SynapseSqlClient(
            dataset_name,
            staging_dataset_name,
            config.credentials,
            capabilities,
        )

        self.active_hints = deepcopy(HINT_TO_SYNAPSE_ATTR)
        if not self.config.create_indexes:
            self.active_hints.pop("primary_key", None)
            self.active_hints.pop("unique", None)

    def _get_table_update_sql(
        self, table_name: str, new_columns: Sequence[TColumnSchema], generate_alter: bool
    ) -> List[str]:
        table = self.prepare_load_table(table_name)
        if self.in_staging_dataset_mode:
            # Staging tables should always be heap tables, because "when you are
            # temporarily landing data in dedicated SQL pool, you may find that
            # using a heap table makes the overall process faster."
            table[TABLE_INDEX_TYPE_HINT] = "heap"  # type: ignore[typeddict-unknown-key]

        table_index_type = cast(TTableIndexType, table.get(TABLE_INDEX_TYPE_HINT))
        if self.in_staging_dataset_mode:
            final_table = self.prepare_load_table(table_name)
            final_table_index_type = cast(TTableIndexType, final_table.get(TABLE_INDEX_TYPE_HINT))
        else:
            final_table_index_type = table_index_type
        if final_table_index_type == "clustered_columnstore_index":
            # Even if the staging table has index type "heap", we still adjust
            # the column data types to prevent errors when writing into the
            # final table that has index type "clustered_columnstore_index".
            new_columns = self._get_columstore_valid_columns(table, new_columns)

        _sql_result = SqlJobClientBase._get_table_update_sql(
            self, table_name, new_columns, generate_alter
        )
        if not generate_alter:
            table_index_type_attr = TABLE_INDEX_TYPE_TO_SYNAPSE_ATTR[table_index_type]
            sql_result = [_sql_result[0] + f"\n WITH ( {table_index_type_attr} );"]
        else:
            sql_result = _sql_result
        return sql_result

    def _get_columstore_valid_columns(
        self, table: PreparedTableSchema, new_columns: Sequence[TColumnSchema]
    ) -> Sequence[TColumnSchema]:
        return [self._get_columstore_valid_column(table, c) for c in new_columns]

    def _get_columstore_valid_column(
        self, table: PreparedTableSchema, c: TColumnSchema
    ) -> TColumnSchema:
        """
        Returns TColumnSchema that maps to a Synapse data type that can participate in a columnstore index.

        varchar(max), nvarchar(max), and varbinary(max) are replaced with
        varchar(n), nvarchar(n), and varbinary(n), respectively, where
        n equals the user-specified precision, or the maximum allowed
        value if the user did not specify a precision.
        """
        dbt = self.capabilities.get_type_mapper().to_destination_type(c, table)
        if "(max)" in dbt and "precision" not in c:
            if c["data_type"] in ("text", "json"):
                return {**c, **{"precision": VARCHAR_MAX_N}}
            elif c["data_type"] in ("binary"):
                return {**c, **{"precision": VARBINARY_MAX_N}}
            else:
                raise ValueError(
                    dbt, f"Unexpected database type `{dbt}` on dlt type `{c['data_type']}`"
                )

        return c

    def _create_replace_followup_jobs(
        self, table_chain: Sequence[PreparedTableSchema]
    ) -> List[FollowupJobRequest]:
        return SqlJobClientBase._create_replace_followup_jobs(self, table_chain)

    def prepare_load_table(self, table_name: str) -> PreparedTableSchema:
        table = super().prepare_load_table(table_name)
        if table_name in self.schema.dlt_table_names():
            # dlt tables should always be heap tables, because "for small lookup
            # tables, less than 60 million rows, consider using HEAP or clustered
            # index for faster query performance."
            table[TABLE_INDEX_TYPE_HINT] = "heap"  # type: ignore[typeddict-unknown-key]
        # https://learn.microsoft.com/en-us/azure/synapse-analytics/sql-data-warehouse/sql-data-warehouse-tables-index#heap-tables
        else:
            if TABLE_INDEX_TYPE_HINT not in table:
                # If present in parent table, fetch hint from there.
                table[TABLE_INDEX_TYPE_HINT] = get_inherited_table_hint(  # type: ignore[typeddict-unknown-key]
                    self.schema.tables, table_name, TABLE_INDEX_TYPE_HINT, allow_none=True
                )
        if table[TABLE_INDEX_TYPE_HINT] is None:  # type: ignore[typeddict-item]
            # Hint still not defined, fall back to default.
            table[TABLE_INDEX_TYPE_HINT] = self.config.default_table_index_type  # type: ignore[typeddict-unknown-key]
        return table

    def create_load_job(
        self, table: PreparedTableSchema, file_path: str, load_id: str, restore: bool = False
    ) -> LoadJob:
        job = super().create_load_job(table, file_path, load_id, restore)
        if not job:
            assert ReferenceFollowupJobRequest.is_reference_job(
                file_path
            ), "Synapse must use staging to load files"
            job = SynapseCopyFileLoadJob(
                file_path,
                self.config.staging_config.credentials,  # type: ignore[arg-type]
                self.config.staging_use_msi,
            )
        return job

    def should_truncate_table_before_load_on_staging_destination(self, table_name: str) -> bool:
        return self.config.truncate_tables_on_staging_destination_before_load


class SynapseCopyFileLoadJob(CopyRemoteFileLoadJob):
    def __init__(
        self,
        file_path: str,
        staging_credentials: Optional[
            Union[AzureCredentialsWithoutDefaults, AzureServicePrincipalCredentialsWithoutDefaults]
        ] = None,
        staging_use_msi: bool = False,
    ) -> None:
        self.staging_use_msi = staging_use_msi
        super().__init__(file_path, staging_credentials)

    def run(self) -> None:
        self._sql_client = self._job_client.sql_client
        # get format
        ext = os.path.splitext(self._bucket_path)[1][1:]
        if ext == "parquet":
            file_type = "PARQUET"

            # dlt-generated DDL statements will still create the table, but
            # enabling AUTO_CREATE_TABLE prevents a MalformedInputException.
            auto_create_table = "ON"
        else:
            raise ValueError(f"Unsupported file type `{ext}` for Synapse.")

        staging_credentials = self._staging_credentials
        assert staging_credentials is not None
        assert isinstance(
            staging_credentials,
            (AzureCredentialsWithoutDefaults, AzureServicePrincipalCredentialsWithoutDefaults),
        )
        azure_storage_account_name = staging_credentials.azure_storage_account_name
        https_path = self._get_https_path(self._bucket_path, azure_storage_account_name)
        table_name = self._load_table["name"]

        if self.staging_use_msi:
            credential = "IDENTITY = 'Managed Identity'"
        else:
            # re-use staging credentials for copy into Synapse
            if isinstance(staging_credentials, AzureCredentialsWithoutDefaults):
                sas_token = staging_credentials.azure_storage_sas_token
                credential = f"IDENTITY = 'Shared Access Signature', SECRET = '{sas_token}'"
            elif isinstance(staging_credentials, AzureServicePrincipalCredentialsWithoutDefaults):
                tenant_id = staging_credentials.azure_tenant_id
                endpoint = f"https://login.microsoftonline.com/{tenant_id}/oauth2/token"
                identity = f"{staging_credentials.azure_client_id}@{endpoint}"
                secret = staging_credentials.azure_client_secret
                credential = f"IDENTITY = '{identity}', SECRET = '{secret}'"
            else:
                raise ConfigurationException(
                    f"Credentials of type `{type(staging_credentials)}` not supported"
                    " when loading data from staging into Synapse using `COPY INTO`."
                )

        # Copy data from staging file into Synapse table.
        with self._sql_client.begin_transaction():
            dataset_name = self._sql_client.dataset_name
            sql = dedent(f"""
                COPY INTO [{dataset_name}].[{table_name}]
                FROM '{https_path}'
                WITH (
                    FILE_TYPE = '{file_type}',
                    CREDENTIAL = ({credential}),
                    AUTO_CREATE_TABLE = '{auto_create_table}'
                )
            """)
            self._sql_client.execute_sql(sql)

    def _get_https_path(self, bucket_path: str, storage_account_name: str) -> str:
        """
        Converts a path in the form of az://<container_name>/<path> to
        https://<storage_account_name>.blob.core.windows.net/<container_name>/<path>
        as required by Synapse.
        """
        bucket_url = urlparse(bucket_path)
        # "blob" endpoint has better performance than "dfs" endoint
        # https://learn.microsoft.com/en-us/sql/t-sql/statements/copy-into-transact-sql?view=azure-sqldw-latest#external-locations
        endpoint = "blob"
        _path = "/" + bucket_url.netloc + bucket_url.path
        https_url = bucket_url._replace(
            scheme="https",
            netloc=f"{storage_account_name}.{endpoint}.core.windows.net",
            path=_path,
        )
        return urlunparse(https_url)



"""
Cached helper methods for all operations that are called often
"""
from functools import lru_cache
from typing import Any, Dict, List, Optional, Tuple, cast

from dlt.common.json import json
from dlt.common.destination.utils import resolve_merge_strategy
from dlt.common.normalizers.json.typing import RelationalNormalizerConfigPropagation
from dlt.common.normalizers.naming import NamingConvention
from dlt.common.normalizers.typing import TRowIdType
from dlt.common.normalizers.utils import DLT_ID_LENGTH_BYTES
from dlt.common.schema import Schema
from dlt.common.schema.typing import (
    TColumnName,
    TColumnSchema,
    C_DLT_ID,
    DLT_NAME_PREFIX,
    TTableSchema,
)
from dlt.common.schema.utils import (
    get_columns_names_with_prop,
    get_first_column_name_with_prop,
    get_nested_tables,
    has_column_with_prop,
    is_nested_table,
)
from dlt.common.utils import digest128, digest128b


def shorten_fragments(naming: NamingConvention, *idents: str) -> str:
    return naming.shorten_fragments(*idents)


def normalize_table_identifier(schema: Schema, naming: NamingConvention, table_name: str) -> str:
    if schema._normalizers_config.get("use_break_path_on_normalize", True):
        return naming.normalize_tables_path(table_name)
    else:
        return naming.normalize_table_identifier(table_name)


def normalize_identifier(schema: Schema, naming: NamingConvention, identifier: str) -> str:
    if schema._normalizers_config.get("use_break_path_on_normalize", True):
        return naming.normalize_path(identifier)
    else:
        return naming.normalize_identifier(identifier)


def get_table_nesting_level(
    schema: Schema, table_name: str, default_nesting: int = 1000
) -> Optional[int]:
    """gets table nesting level, will inherit from parent if not set"""

    table = schema.tables.get(table_name)
    if (
        table
        and (max_nesting := cast(int, table.get("x-normalizer", {}).get("max_nesting"))) is not None
    ):
        return max_nesting
    return default_nesting


def get_primary_key(schema: Schema, table_name: str) -> List[str]:
    if table_name not in schema.tables:
        return []
    table = schema.get_table(table_name)
    pk = get_columns_names_with_prop(table, "primary_key", include_incomplete=True)
    return pk


def is_nested_type(
    schema: Schema,
    table_name: str,
    field_name: str,
    _r_lvl: int,
) -> bool:
    """For those paths the nested objects should be left in place.
    Cache perf: max_nesting < _r_lvl: ~2x faster, full check 10x faster
    """

    # nesting level is counted backwards
    # is we have traversed to or beyond the calculated nesting level, we detect a nested type
    if _r_lvl <= 0:
        return True

    column: TColumnSchema = None
    table = schema.tables.get(table_name)
    if table:
        column = table["columns"].get(field_name)

    # JSON expansion hints are handled in the expansion layer, not by nested-type detection.
    # If a column defines any x-json-* hint (value not None), treat it as non-nested here.
    if column is not None and (
        column.get("x-json-flatten") is not None
        or column.get("x-json-keep-original") is not None
    ):
        return False

    if column is None or "data_type" not in column:
        data_type = schema.get_preferred_type(field_name)
    else:
        data_type = column["data_type"]

    return data_type == "json"


def should_be_nested(schema: Schema, table_name: str) -> bool:
    """Tells if table should be nested or should be a root table. All tables created by the normalizer
    are nested, defined tables are checked.
    """
    if table := schema.tables.get(table_name):
        return is_nested_table(table)
    return True


def get_root_row_id_type(schema: Schema, table_name: str) -> TRowIdType:
    if table := schema.tables.get(table_name):
        merge_strategy = resolve_merge_strategy(schema.tables, table)
        if merge_strategy == "upsert":
            return "key_hash"
        elif merge_strategy == "scd2":
            x_row_version_col = get_first_column_name_with_prop(
                schema.get_table(table_name),
                "x-row-version",
                include_incomplete=True,
            )
            if x_row_version_col == schema.naming.normalize_identifier(C_DLT_ID):
                return "row_hash"
    return "random"


def get_propagation_mapping(
    config: RelationalNormalizerConfigPropagation, table: str, is_root: bool
) -> Dict[TColumnName, TColumnName]:
    # mapping(k:v): propagate property with name "k" as property with name "v" in nested table
    mappings = {}
    if is_root:
        mappings.update(config.get("root") or {})
    if table in (config.get("tables") or {}):
        mappings.update(config["tables"][table])
    return mappings


def requires_root_key(
    schema: Schema, root_table: TTableSchema, root_key_propagation: Optional[bool]
) -> bool:
    """Checks if table chain containing `table` requires root_key to propagate.

    1. if there's any table with `root_key` in table chain already - we always propagate
    2. if write disposition is merge and merge strategy is "delete-insert", "upsert" and root_key_propagation is not False
    3. root_key_propagation is True
    """
    table_name = root_table["name"]
    assert not is_nested_table(root_table), f"{table_name} cannot be nested table"
    if table_name in schema.dlt_table_names():
        merge_requires = False
    else:
        merge_strategy = resolve_merge_strategy(schema.tables, root_table)
        merge_requires = (
            merge_strategy in ["delete-insert", "upsert"]
            if root_key_propagation is None
            else root_key_propagation
        )
    return merge_requires or any(
        has_column_with_prop(t, "root_key", include_incomplete=True)
        for t in get_nested_tables(schema.tables, table_name)
    )


def get_row_hash(row: Dict[str, Any], subset: Optional[List[str]] = None) -> str:
    """Returns hash of row.

    Hash includes column names and values and is ordered by column name.
    Excludes dlt system columns.
    Can be used as deterministic row identifier.
    """
    if subset is not None:
        # all parts of the key must be present
        row_filtered = {k: row[k] for k in subset}
    else:
        row_filtered = {k: v for k, v in row.items() if not k.startswith(DLT_NAME_PREFIX)}
    row_str = json.dumpb(row_filtered, sort_keys=True)
    return digest128b(row_str, DLT_ID_LENGTH_BYTES)


def get_nested_row_hash(parent_row_id: str, nested_table: str, list_idx: int) -> str:
    # create deterministic unique id of the nested row taking into account that all lists are ordered
    # and all nested tables must be lists
    return digest128(f"{parent_row_id}_{nested_table}_{list_idx}", DLT_ID_LENGTH_BYTES)



from typing import Any, Dict, Union, Literal

from dlt.common.exceptions import PipelineException
from dlt.common.pipeline import LoadInfo, NormalizeInfo, StepInfo, StepMetrics, SupportsPipeline

from dlt.common.storages.load_package import PackageStorage
from dlt.pipeline.typing import TPipelineStep


class InvalidPipelineName(PipelineException, ValueError):
    def __init__(self, pipeline_name: str, details: str) -> None:
        super().__init__(
            pipeline_name,
            f"The pipeline name `{pipeline_name}` contains invalid characters. The pipeline name is"
            " used to create a pipeline working directory and must be a valid directory name. "
            f"Details: {details}",
        )


class PipelineConfigMissing(PipelineException):
    def __init__(
        self,
        pipeline_name: str,
        config_elem: str,
        step_or_func: Union[TPipelineStep, Literal["dataset"]],
        _help: str = None,
    ) -> None:
        self.config_elem = config_elem
        self.step_or_func = step_or_func
        msg = (
            f"Configuration element `{config_elem}` was not provided and step or function"
            f" `{step_or_func}` cannot be executed"
        )
        if _help:
            msg += f"\n{_help}\n"
        super().__init__(pipeline_name, msg)


class CannotRestorePipelineException(PipelineException):
    def __init__(self, pipeline_name: str, pipelines_dir: str, reason: str) -> None:
        msg = (
            f"Pipeline with `{pipeline_name=:}` and `{pipelines_dir=:}` could not be"
            f" restored: {reason}"
        )
        super().__init__(pipeline_name, msg)


class PipelineStepFailed(PipelineException):
    """Raised by run, extract, normalize and load Pipeline methods."""

    def __init__(
        self,
        pipeline: SupportsPipeline,
        step: TPipelineStep,
        load_id: str,
        exception: BaseException,
        step_info: StepInfo[StepMetrics] = None,
    ) -> None:
        self.pipeline = pipeline
        self.step = step
        self.load_id = load_id
        self.exception = exception
        self.step_info = step_info
        self.has_pending_data = pipeline.has_pending_data
        self.is_package_partially_loaded = False

        package_str = f" when processing package with `{load_id=:}`" if load_id else ""
        msg = (
            f"Pipeline execution failed at `{step=:}`{package_str} with"
            f" exception:\n\n{type(exception)}\n{exception}"
        )
        if isinstance(step_info, (NormalizeInfo, LoadInfo)):
            if self.has_pending_data:
                msg += (
                    "\n\nPending packages are left in the pipeline and will be re-tried on the"
                    " next pipeline run."
                    " If you pass new data to extract to next run, it will be ignored. Run "
                    f"`dlt pipeline {pipeline.pipeline_name} info` for more information or `dlt"
                    f" pipeline {pipeline.pipeline_name} drop-pending-packages` to drop pending"
                    " packages."
                )
            if load_id and step_info and load_id in step_info.loads_ids and step == "load":
                # get package info
                package_info = next(
                    (p for p in step_info.load_packages if p.load_id == load_id), None
                )
                if package_info:
                    self.is_package_partially_loaded = PackageStorage.is_package_partially_loaded(
                        package_info
                    )
                    if self.is_package_partially_loaded:
                        msg += (
                            f"\nWARNING: package `{load_id}` is partially loaded. Data in"
                            " destination could be modified by one of completed load jobs while"
                            " others were not yet executed or were retried. Data in the"
                            " destination may be in inconsistent state. We recommend that you"
                            " retry the load or  review the incident before dropping pending"
                            " packages. See"
                            " https://dlthub.com/docs/running-in-production/running#partially-loaded-packages"
                            " for details"
                        )

        super().__init__(pipeline.pipeline_name, msg)

    def attrs(self) -> Dict[str, Any]:
        # remove attr that should not be published
        attrs_ = super().attrs()
        attrs_.pop("pipeline")
        attrs_.pop("exception")
        attrs_.pop("step_info")
        return attrs_


class PipelineStateEngineNoUpgradePathException(PipelineException):
    def __init__(
        self, pipeline_name: str, init_engine: int, from_engine: int, to_engine: int
    ) -> None:
        self.init_engine = init_engine
        self.from_engine = from_engine
        self.to_engine = to_engine
        super().__init__(
            pipeline_name,
            f"No engine upgrade path for state in pipeline `{pipeline_name}` from `{init_engine}`"
            f" to `{to_engine}`, stopped at `{from_engine}`. You possibly tried to run an older dlt"
            " version against a destination you have previously loaded data to with a newer dlt"
            " version.",
        )


class PipelineHasPendingDataException(PipelineException):
    def __init__(self, pipeline_name: str, pipelines_dir: str) -> None:
        msg = (
            f" Operation failed because pipeline with `{pipeline_name=:}` and `{pipelines_dir=:}`"
            " contains pending extracted files or load packages. Use `dlt pipeline sync`"
            " to reset the local state then run this operation again."
        )
        super().__init__(pipeline_name, msg)


class PipelineNeverRan(PipelineException):
    def __init__(self, pipeline_name: str, pipelines_dir: str) -> None:
        msg = (
            f" Operation failed because pipeline with `{pipeline_name=:}` and `{pipelines_dir=:}`"
            " was never run or was never synced with destination. Use `dlt pipeline sync`"
            " or call `pipeline.sync_destination()` to get pipeline state and set of"
            " schemas from destination."
        )
        super().__init__(pipeline_name, msg)


class PipelineNotActive(PipelineException):
    def __init__(self, pipeline_name: str) -> None:
        super().__init__(
            pipeline_name, f"Pipeline `{pipeline_name}` is not active so it cannot be deactivated"
        )



import inspect
from typing import (
    Callable,
    Optional,
    Tuple,
    Union,
    List,
    Any,
    Sequence,
    cast,
    AsyncIterator,
    Awaitable,
    Generator,
    Iterator,
)
from collections.abc import Mapping as C_Mapping
from functools import wraps

from dlt.common.data_writers import TDataItemFormat
from dlt.common.exceptions import MissingDependencyException
from dlt.common.reflection.inspect import isgeneratorfunction
from dlt.common.schema.typing import TAnySchemaColumns, TTableSchemaColumns
from dlt.common.schema.utils import normalize_schema_name
from dlt.common.typing import (
    Self,
    AnyFun,
    DictStrAny,
    TDataItem,
    TDataItems,
    TAnyFunOrGenerator,
    TColumnNames,
    NoneType,
)
from dlt.common.utils import get_callable_name

from dlt.extract.exceptions import (
    InvalidResourceDataTypeIsNone,
    InvalidResourceReturnsResource,
    InvalidStepFunctionArguments,
)
from dlt.extract.items import (
    TTableHintTemplate,
    TFunHintTemplate,
    SupportsPipe,
)

from dlt.common.schema.typing import TFileFormat

try:
    from dlt.common.libs import pydantic
except MissingDependencyException:
    pydantic = None


try:
    from dlt.common.libs import pyarrow
except MissingDependencyException:
    pyarrow = None

try:
    from dlt.common.libs.pandas import pandas
except MissingDependencyException:
    pandas = None


def get_data_item_format(items: TDataItems) -> TDataItemFormat:
    """Detect the format of the data item from `items`.

    Reverts to `object` for empty lists

    Returns:
        The data file format.
    """

    # if incoming item is hints meta, check if item format is forced
    from dlt.destinations.dataset.relation import ReadableDBAPIRelation

    if isinstance(items, ReadableDBAPIRelation):
        return "model"

    if not pyarrow and not pandas:
        return "object"

    # Assume all items in list are the same type
    try:
        if isinstance(items, list):
            items = items[0]
        if (pyarrow and pyarrow.is_arrow_item(items)) or (
            pandas and isinstance(items, pandas.DataFrame)
        ):
            return "arrow"
    except IndexError:
        pass
    return "object"


def resolve_column_value(
    column_hint: TTableHintTemplate[TColumnNames], item: TDataItem
) -> Union[Any, List[Any]]:
    """Extract values from the data item given a column hint.
    Returns either a single value or list of values when hint is a composite.
    """
    columns = column_hint(item) if callable(column_hint) else column_hint
    if isinstance(columns, str):
        return item[columns]
    return [item[k] for k in columns]


def ensure_table_schema_columns(columns: TAnySchemaColumns) -> TTableSchemaColumns:
    """Convert supported column schema types to a column dict which
    can be used in resource schema.

    Args:
        columns: A dict of column schemas, a list of column schemas, or a pydantic model
    """
    if isinstance(columns, C_Mapping):
        # fill missing names in short form was used
        for col_name in columns:
            columns[col_name]["name"] = col_name
        return columns
    elif isinstance(columns, Sequence):
        # Assume list of columns
        return {col["name"]: col for col in columns}
    elif pydantic is not None and (
        isinstance(columns, pydantic.BaseModel) or issubclass(columns, pydantic.BaseModel)
    ):
        return pydantic.pydantic_to_table_schema_columns(columns)

    raise ValueError(f"Unsupported columns type: `{type(columns)}`")


def ensure_table_schema_columns_hint(
    columns: TTableHintTemplate[TAnySchemaColumns],
) -> TTableHintTemplate[TTableSchemaColumns]:
    """Convert column schema hint to a hint returning `TTableSchemaColumns`.
    A callable hint is wrapped in another function which converts the original result.
    """
    if callable(columns) and not isinstance(columns, type):

        def wrapper(item: TDataItem) -> TTableSchemaColumns:
            return ensure_table_schema_columns(
                cast(TFunHintTemplate[TAnySchemaColumns], columns)(item)
            )

        return wrapper

    return ensure_table_schema_columns(columns)


def simulate_func_call(
    f: Union[Any, AnyFun], args_to_skip: int, *args: Any, **kwargs: Any
) -> Tuple[inspect.Signature, inspect.Signature, inspect.BoundArguments]:
    """Simulates a call to a resource or transformer function before it will be wrapped for later execution in the pipe

    Returns a tuple with a `f` signature, modified signature in case of transformers and bound arguments
    """
    if not callable(f):
        # just provoke a call to raise default exception
        f()
    assert callable(f)

    sig = inspect.signature(f)
    # simulate the call to the underlying callable
    no_item_sig = sig.replace(parameters=list(sig.parameters.values())[args_to_skip:])
    try:
        bound_args = no_item_sig.bind(*args, **kwargs)
    except TypeError as v_ex:
        raise TypeError(f"{get_callable_name(f)}(): " + str(v_ex))
    return sig, no_item_sig, bound_args


def check_compat_transformer(name: str, f: AnyFun, sig: inspect.Signature) -> inspect.Parameter:
    sig_arg_count = len(sig.parameters)
    callable_name = get_callable_name(f)
    if sig_arg_count == 0:
        raise InvalidStepFunctionArguments(name, callable_name, sig, "Function takes no arguments")

    # see if meta is present in kwargs
    meta_arg = next((p for p in sig.parameters.values() if p.name == "meta"), None)
    if meta_arg is not None:
        if meta_arg.kind not in (meta_arg.KEYWORD_ONLY, meta_arg.POSITIONAL_OR_KEYWORD):
            raise InvalidStepFunctionArguments(
                name, callable_name, sig, "'meta' cannot be pos only argument '"
            )
    return meta_arg


def wrap_iterator(gen: Iterator[TDataItems]) -> Iterator[TDataItems]:
    """Wraps an iterator into a generator"""
    if inspect.isgenerator(gen):
        return gen

    def wrapped_gen() -> Iterator[TDataItems]:
        yield from gen

    return wrapped_gen()


def wrap_async_iterator(
    gen: AsyncIterator[TDataItems],
) -> Generator[Awaitable[TDataItems], None, None]:
    """Wraps an async generator into a list of awaitables"""
    exhausted = False
    busy = False

    # creates an awaitable that will return the next item from the async generator
    async def run() -> TDataItems:
        nonlocal exhausted
        try:
            # if marked exhausted by the main thread and we are wrapping a generator
            # we can close it here
            if exhausted:
                raise StopAsyncIteration()
            item = await gen.__anext__()
            return item
        # on stop iteration mark as exhausted
        # also called when futures are cancelled
        except StopAsyncIteration:
            exhausted = True
            raise
        finally:
            nonlocal busy
            busy = False

    # this generator yields None while the async generator is not exhausted
    try:
        while not exhausted:
            while busy:
                yield None
            busy = True
            yield run()
    # this gets called from the main thread when the wrapping generater is closed
    except GeneratorExit:
        # mark as exhausted
        exhausted = True


def wrap_parallel_iterator(f: TAnyFunOrGenerator) -> TAnyFunOrGenerator:
    """Wraps a generator for parallel extraction"""

    def _gen_wrapper(*args: Any, **kwargs: Any) -> Iterator[TDataItems]:
        gen: TAnyFunOrGenerator
        if callable(f):
            gen = f(*args, **kwargs)
        else:
            gen = f

        exhausted = False
        busy = False

        def _parallel_gen() -> TDataItems:
            nonlocal busy
            nonlocal exhausted
            try:
                return next(gen)  # type: ignore[call-overload]
            except StopIteration:
                exhausted = True
                return None
            finally:
                busy = False

        while not exhausted:
            try:
                while busy:
                    yield None
                busy = True
                yield _parallel_gen
            except GeneratorExit:
                gen.close()  # type: ignore[attr-defined]
                raise

    if callable(f):
        if isgeneratorfunction(f):
            return wraps(f)(_gen_wrapper)  # type: ignore[return-value]
        else:

            def _fun_wrapper(*args: Any, **kwargs: Any) -> Any:
                def _curry() -> Any:
                    return f(*args, **kwargs)

                return _curry

            return wraps(f)(_fun_wrapper)  # type: ignore[return-value]
    return _gen_wrapper()  # type: ignore[return-value]


def _transformer_compat(item: TDataItems, meta: Any = None) -> Any:
    pass


_transformer_compat_sig = inspect.signature(_transformer_compat)


def wrap_compat_transformer(
    name: str, f: AnyFun, sig: inspect.Signature, *args: Any, **kwargs: Any
) -> AnyFun:
    """Creates a compatible wrapper over transformer function. A pure transformer function expects data item in first argument and one keyword argument called `meta`"""
    check_compat_transformer(name, f, sig)
    if len(sig.parameters) == 2 and "meta" in sig.parameters:
        return f

    def _tx_partial(item: TDataItems, meta: Any = None) -> Any:
        # print(f"_ITEM:{item}{meta},{args}{kwargs}")
        # also provide optional meta so pipe does not need to update arguments
        if "meta" in kwargs:
            kwargs["meta"] = meta
        return f(item, *args, **kwargs)

    # this partial wraps transformer and sets a signature that is compatible with pipe transform calls
    _wrapper = wraps(f)(_tx_partial)
    setattr(_wrapper, "__signature__", _transformer_compat_sig)
    return _wrapper


def _resource_compat() -> Any:
    pass


_resource_compat_sig = inspect.signature(_resource_compat)


def wrap_resource_gen(
    name: str, f: AnyFun, sig: inspect.Signature, *args: Any, **kwargs: Any
) -> AnyFun:
    """Wraps a generator or generator function so it is evaluated on extraction and binds argument passed to call"""

    def _partial() -> Any:
        # print(f"_PARTIAL: {args} {kwargs}")
        rv = f(*args, **kwargs)
        if rv is None:
            raise InvalidResourceDataTypeIsNone(name, rv, NoneType)
        # is it Pipe or resource
        if hasattr(rv, "_gen_idx") or hasattr(rv, "_pipe"):
            raise InvalidResourceReturnsResource(name, f, type(f))
        return rv

    _wrapper = wraps(f)(_partial)
    setattr(_wrapper, "__signature__", _resource_compat_sig)
    return _wrapper


def make_schema_with_default_name(pipeline_name: str) -> str:
    """Makes schema name from the pipeline name using the name normalizer. "_pipeline" suffix is removed if present"""
    if pipeline_name.endswith("_pipeline"):
        schema_name = pipeline_name[:-9]
    else:
        schema_name = pipeline_name
    return normalize_schema_name(schema_name)


class dynstr(str):
    """Dynamic string which will generate final value when called"""

    __slots__ = "dyn"

    def __new__(cls, placeholder: str, dyn: Callable[..., str]) -> Self:
        # Create a new str instance
        instance = super().__new__(cls, placeholder)
        return instance

    def __init__(self, placeholder: str, dyn: Callable[..., str]):
        super().__init__()
        self.dyn = dyn

    def __call__(self, param: Any) -> str:
        return self.dyn(param)



from email.utils import parsedate_tz, mktime_tz
import re
import time
from typing import (
    Optional,
    cast,
    Callable,
    Type,
    Union,
    Sequence,
    Tuple,
    List,
    TYPE_CHECKING,
    Any,
    Dict,
)
from threading import local

from requests import Response, HTTPError, Session as BaseSession
from requests.exceptions import ConnectionError, Timeout, ChunkedEncodingError
from requests.adapters import HTTPAdapter
from tenacity import (
    Retrying,
    retry_if_exception_type,
    stop_after_attempt,
    RetryCallState,
    retry_any,
    wait_exponential,
)
from tenacity.retry import retry_base

from dlt.common.configuration.inject import with_config
from dlt.common.typing import TimedeltaSeconds, ConfigValue
from dlt.common.configuration.specs import RuntimeConfiguration
from dlt.sources.helpers.requests.session import Session, DEFAULT_TIMEOUT
from dlt.sources.helpers.requests.typing import TRequestTimeout


DEFAULT_RETRY_STATUS = (429, *range(500, 600))
DEFAULT_RETRY_EXCEPTIONS = (ConnectionError, Timeout, ChunkedEncodingError)

RetryPredicate = Callable[[Optional[Response], Optional[BaseException]], bool]


def _get_retry_response(retry_state: RetryCallState) -> Optional[Response]:
    ex = retry_state.outcome.exception()
    if ex:
        if isinstance(ex, HTTPError):
            return ex.response
        return None
    result = retry_state.outcome.result()
    return result if isinstance(result, Response) else None


class retry_if_status(retry_base):
    """Retry for given response status codes"""

    def __init__(self, status_codes: Sequence[int]) -> None:
        self.status_codes = set(status_codes)

    def __call__(self, retry_state: RetryCallState) -> bool:
        response = _get_retry_response(retry_state)
        if response is None:
            return False
        result = response.status_code in self.status_codes
        return result


class retry_if_predicate(retry_base):
    def __init__(self, predicate: RetryPredicate) -> None:
        self.predicate = predicate

    def __call__(self, retry_state: RetryCallState) -> bool:
        response = _get_retry_response(retry_state)
        exception = retry_state.outcome.exception()
        return self.predicate(response, exception)


class wait_exponential_retry_after(wait_exponential):
    def _parse_retry_after(self, retry_after: str) -> Optional[float]:
        # Borrowed from urllib3
        seconds: float
        # Whitespace: https://tools.ietf.org/html/rfc7230#section-3.2.4
        if re.match(r"^\s*[0-9]+\s*$", retry_after):
            seconds = int(retry_after)
        else:
            retry_date_tuple = parsedate_tz(retry_after)
            if retry_date_tuple is None:
                return None
            retry_date = mktime_tz(retry_date_tuple)
            seconds = retry_date - time.time()
        return max(self.min, min(self.max, seconds))

    def _get_retry_after(self, retry_state: RetryCallState) -> Optional[float]:
        response = _get_retry_response(retry_state)
        if response is None:
            return None
        header = response.headers.get("Retry-After")
        if not header:
            return None
        return self._parse_retry_after(header)

    def __call__(self, retry_state: RetryCallState) -> float:
        retry_after = self._get_retry_after(retry_state)
        if retry_after is not None:
            return retry_after
        return super().__call__(retry_state)


def _make_retry(
    status_codes: Sequence[int],
    exceptions: Sequence[Type[Exception]],
    max_attempts: int,
    condition: Union[RetryPredicate, Sequence[RetryPredicate], None],
    backoff_factor: float,
    respect_retry_after_header: bool,
    max_delay: TimedeltaSeconds,
) -> Retrying:
    retry_conds = [retry_if_status(status_codes), retry_if_exception_type(tuple(exceptions))]
    if condition is not None:
        if callable(condition):
            condition = [condition]
        retry_conds.extend([retry_if_predicate(c) for c in condition])

    wait_cls = wait_exponential_retry_after if respect_retry_after_header else wait_exponential
    return Retrying(
        wait=wait_cls(multiplier=backoff_factor, max=max_delay),
        retry=(retry_any(*retry_conds)),
        stop=stop_after_attempt(max_attempts),
        reraise=True,
        retry_error_callback=lambda state: state.outcome.result(),
    )


class Client:
    """Wrapper for `requests` to create a `Session` with configurable retry functionality.

    #### Note:
    Create a  `requests.Session` which automatically retries requests in case of error.
    By default retries are triggered for `5xx` and `429` status codes and when the server is unreachable or drops connection.

    #### Custom retry condition
    You can provide one or more custom predicates for specific retry condition. The predicate is called after every request with the resulting response and/or exception.
    For example, this will trigger a retry when the response text is `error`:

    >>> from typing import Optional
    >>> from requests import Response
    >>>
    >>> def should_retry(response: Optional[Response], exception: Optional[BaseException]) -> bool:
    >>>     if response is None:
    >>>         return False
    >>>     return response.text == 'error'

    The retry is triggered when either any of the predicates or the default conditions based on status code/exception are `True`.

    Args:
        request_timeout: Timeout for requests in seconds. May be passed as `timedelta` or `float/int` number of seconds.
        max_connections: Max connections per host in the HTTPAdapter pool
        raise_for_status: Whether to raise exception on error status codes (using `response.raise_for_status()`)
        session: Optional `requests.Session` instance to add the retry handler to. A new session is created by default.
        status_codes: Retry when response has any of these status codes. Default `429` and all `5xx` codes. Pass an empty list to disable retry based on status.
        exceptions: Retry on exception of given type(s). Default `(requests.Timeout, requests.ConnectionError)`. Pass an empty list to disable retry on exceptions.
        request_max_attempts: Max number of retry attempts before giving up
        retry_condition: A predicate or a list of predicates to decide whether to retry. If any predicate returns `True` the request is retried
        request_backoff_factor: Multiplier used for exponential delay between retries
        request_max_retry_delay: Maximum delay when using exponential backoff
        respect_retry_after_header: Whether to use the `Retry-After` response header (when available) to determine the retry delay
        session_attrs: Extra attributes that will be set on the session instance, e.g. `{headers: {'Authorization': 'api-key'}}` (see `requests.sessions.Session` for possible attributes)
    """

    _session_attrs: Dict[str, Any]

    def __init__(
        self,
        request_timeout: Optional[
            Union[TimedeltaSeconds, Tuple[TimedeltaSeconds, TimedeltaSeconds]]
        ] = DEFAULT_TIMEOUT,
        max_connections: int = 50,
        raise_for_status: bool = True,
        status_codes: Sequence[int] = DEFAULT_RETRY_STATUS,
        exceptions: Sequence[Type[Exception]] = DEFAULT_RETRY_EXCEPTIONS,
        request_max_attempts: int = RuntimeConfiguration.request_max_attempts,
        retry_condition: Union[RetryPredicate, Sequence[RetryPredicate], None] = None,
        request_backoff_factor: float = RuntimeConfiguration.request_backoff_factor,
        request_max_retry_delay: TimedeltaSeconds = RuntimeConfiguration.request_max_retry_delay,
        respect_retry_after_header: bool = True,
        session_attrs: Optional[Dict[str, Any]] = None,
    ) -> None:
        self._adapter = HTTPAdapter(pool_maxsize=max_connections)
        self._local = local()
        self._session_kwargs = dict(timeout=request_timeout, raise_for_status=raise_for_status)
        self._retry_kwargs: Dict[str, Any] = dict(
            status_codes=status_codes,
            exceptions=exceptions,
            max_attempts=request_max_attempts,
            condition=retry_condition,
            backoff_factor=request_backoff_factor,
            respect_retry_after_header=respect_retry_after_header,
            max_delay=request_max_retry_delay,
        )
        self._session_attrs = session_attrs or {}

        if TYPE_CHECKING:
            self.get = self.session.get
            self.post = self.session.post
            self.put = self.session.put
            self.patch = self.session.patch
            self.delete = self.session.delete
            self.head = self.session.head
            self.options = self.session.options
            self.request = self.session.request

        self.get = lambda *a, **kw: self.session.get(*a, **kw)
        self.post = lambda *a, **kw: self.session.post(*a, **kw)
        self.put = lambda *a, **kw: self.session.put(*a, **kw)
        self.patch = lambda *a, **kw: self.session.patch(*a, **kw)
        self.delete = lambda *a, **kw: self.session.delete(*a, **kw)
        self.head = lambda *a, **kw: self.session.head(*a, **kw)
        self.options = lambda *a, **kw: self.session.options(*a, **kw)
        self.request = lambda *a, **kw: self.session.request(*a, **kw)

        self._config_version: int = (
            0  # Incrementing marker to ensure per-thread sessions are recreated on config changes
        )

    @with_config(spec=RuntimeConfiguration)
    def configure(self, config: RuntimeConfiguration = ConfigValue) -> None:
        """Update session/retry settings via injected RunConfiguration"""
        self.update_from_config(config)

    def update_from_config(self, config: RuntimeConfiguration) -> None:
        """Update session/retry settings from RunConfiguration"""
        self._session_kwargs["timeout"] = config.request_timeout
        self._retry_kwargs["backoff_factor"] = config.request_backoff_factor
        self._retry_kwargs["max_delay"] = config.request_max_retry_delay
        self._retry_kwargs["max_attempts"] = config.request_max_attempts
        self._config_version += 1

    def _make_session(self) -> Session:
        session = Session(**self._session_kwargs)  # type: ignore[arg-type]
        for key, value in self._session_attrs.items():
            setattr(session, key, value)
        session.mount("http://", self._adapter)
        session.mount("https://", self._adapter)
        retry = _make_retry(**self._retry_kwargs)
        session.send = retry.wraps(session.send)  # type: ignore[method-assign]
        return session

    @property
    def session(self) -> Session:
        session: Optional[Session] = getattr(self._local, "session", None)
        version = self._config_version
        if session is not None:
            version = self._local.config_version
        if session is None or version != self._config_version:
            # Create a new session if config has changed
            session = self._local.session = self._make_session()
            self._local.config_version = self._config_version
        return session
